#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - data
# ==========================================
#
# Get plot data, including isochron diagram data points, age spectra data
#
"""

from math import atan, cos, sin
from . import arr
import numpy as np


def get_data(x: list, sx: list, y: list, sy: list, z: list, sz: list, f: int = 1):
    """
    Get isochron data based on isotopic values. Values of two x and y axes of isochron
    plots are x / z and y / z.

    Parameters
    ----------
    x :
    sx :
    y :
    sy :
    z :
    sz :
    f :

    Returns
    -------
    2D lists  [X, sX, Y, sY, Ri]
        X and Y are coordinate values of data points in x and y axes respectively.
        sX and sY are corresponding errors. Ri is a list of error correlation factors.
    """
    if np.issubdtype(type(f), np.integer) and f > 1:
        sx, sy, sz = np.divide([sx, sy, sz], f)
    # x / z
    k0, k1 = arr.div((x, sx), (z, sz))
    # y / z
    k2, k3 = arr.div((y, sy), (z, sz))
    # ri
    k4 = arr.cor(np.divide(sx, x), np.divide(sy, y), np.divide(sz, z))
    return [k0, k1, k2, k3, k4]


def get_3d_data(x1: list, sx1: list, x2: list, sx2: list,
                x3: list, sx3: list, z: list, sz: list, f: int = 1):
    """ Get values of three axes, x, y, z, based on the given isotopic values
    x = x1 / z, y = x2 / z, z = x3 / z
    Parameters
    ----------
    x1
    sx1
    x2
    sx2
    x3
    sx3
    z
    sz
    f

    Returns
    -------
    list.
    """
    if np.issubdtype(type(f), np.integer) and f > 1:
        sx1, sx2, sx3 = np.divide([sx1, sx2, sx3], f)
    # x1 / z
    k0, k1 = arr.div((x1, sx1), (z, sz))
    # x2 / z
    k2, k3 = arr.div((x2, sx2), (z, sz))
    # x3 / z
    k4, k5 = arr.div((x3, sx3), (z, sz))
    # r1 between x and y
    r1 = arr.cor(np.divide(sx1, x1), np.divide(sx2, x2), np.divide(sz, z))
    # r2 between x and z
    r2 = arr.cor(np.divide(sx1, x1), np.divide(sx3, x3), np.divide(sz, z))
    # r3 between y and z
    r3 = arr.cor(np.divide(sx2, x2), np.divide(sx3, x3), np.divide(sz, z))
    return [k0, k1, k2, k3, k4, k5, r1, r2, r3]


def get_ellipse(x: float, sx: float, y: float, sy: float, r: float, plt_sfactor: int = 1, size: int = 24):
    """ Error ellipses of data points in isochrons. Get 24 points of a ellispse

    Parameters
    ----------
    x
    sx
    y
    sy
    r
    plt_sfactor
    size

    Returns
    -------
    list of 24 points. [[x1, y1], ..., [x24, y24]]
    """
    x, sx, y, sy, r = float(x), float(sx), float(y), float(sy), float(r)
    Qxx = sx ** 2
    Qxy = sx * sy * r
    Qyy = sy ** 2
    # Calculate the ellipse's short semi-axial and long semi-axial
    k = pow((Qxx - Qyy) ** 2 + 4 * Qxy ** 2, 0.5)
    Qee = (Qxx + Qyy + k) / 2
    Qff = (Qxx + Qyy - k) / 2
    e = pow(Qee, 0.5)  # long semi-axial
    f = pow(Qff, 0.5)  # short semi-axial
    phi_e = atan((Qee - Qxx) / Qxy) if Qxy != 0 else 0 if Qxx >= Qyy else np.pi / 2  # radian

    # adjust
    if plt_sfactor == 1:
        v = 2.279  # 68% confidence limit, 1 sigma
    elif plt_sfactor == 2:
        v = 5.991  # 95% confidence limit, 2 sigma, Isoplot R always gives ellipse with 95% confidence
    else:
        v = 1
    e = e * pow(v, 0.5)
    f = f * pow(v, 0.5)

    ellipse_points = []
    for i in range(size):
        theta = i * 2 / size * np.pi
        ellipse_points.append([
            e * cos(theta) * cos(phi_e) - f * sin(theta) * sin(phi_e) + x,
            e * cos(theta) * sin(phi_e) + f * sin(theta) * cos(phi_e) + y
        ])

    return ellipse_points


def get_line_points(xscale, yscale, coeffs):
    """

    Parameters
    ----------
    xscale : x boundary, [min, max]
    yscale : y boundary, [min, max]
    coeffs : y = coeffs[0] + coeffs[1:] * [x]

    Returns
    -------
    List of data points
    """
    xscale = [float(i) for i in xscale]
    yscale = [float(i) for i in yscale]
    coeffs = [float(i) for i in coeffs]
    if not isinstance(coeffs, list) or len(coeffs) < 2:
        raise ValueError(f"Coeffs should be a list with length with 2.")
    get_y = lambda x: coeffs[0] + coeffs[1] * x
    get_x = lambda y: (y - coeffs[0]) / coeffs[1] if coeffs[1] != 0 else None
    res = []
    for point in [
        [xscale[0], get_y(xscale[0])], [xscale[1], get_y(xscale[1])],
        [get_x(yscale[0]), yscale[0]], [get_x(yscale[1]), yscale[1]],
    ]:
        if xscale[0] <= point[0] <= xscale[1] and yscale[0] <= point[1] <= yscale[1]:
            res.append(point)
    return res


def get_set_data(total_data: list, set1_index: list, set2_index: list, unselected_index: list):
    """

    Parameters
    ----------
    total_data
    set1_index
    set2_index
    unselected_index

    Returns
    -------

    """
    set_1, set_2, unslected = [], [], []
    # Remove string and None type in data
    none_index = []
    for each in total_data:
        none_index = none_index + [key for key, val in enumerate(each) if isinstance(val, (str, type(None)))]
    none_index = list(set(none_index))
    none_index.sort(reverse=True)
    for each in total_data:
        for i in none_index:
            each.pop(i)
    set1_index.sort()
    set2_index.sort()
    for col in total_data:
        unslected.append([col[i] for i in unselected_index if i < len(col)])
        set_1.append([col[i] for i in set1_index if i < len(col)])
        set_2.append([col[i] for i in set2_index if i < len(col)])
    return set_1, set_2, unslected



#  Copyright (C) 2024 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2024 Yang 
# ararpy - test_error_correlation.py
# ==========================================
#
#
# 
"""
import numpy as np

ar36 = [float(line.rstrip("\n")) for line in open("ar36.txt").readlines()]
ar37 = [float(line.rstrip("\n")) for line in open("ar37.txt").readlines()]
ar38 = [float(line.rstrip("\n")) for line in open("ar38.txt").readlines()]
ar39 = [float(line.rstrip("\n")) for line in open("ar39.txt").readlines()]
ar40 = [float(line.rstrip("\n")) for line in open("ar40.txt").readlines()]

cov_40_39 = np.cov(ar40, ar39)
cov_40_36 = np.cov(ar40, ar36)
cov_39_36 = np.cov(ar39, ar36)

pho_40_39 = cov_40_39[0][1] / (np.std(ar40) * np.std(ar39))
pho_40_36 = cov_40_36[0][1] / (np.std(ar40) * np.std(ar36))
pho_39_36 = cov_39_36[0][1] / (np.std(ar39) * np.std(ar36))

print(f"{cov_40_39 = }")
print(f"{pho_40_39 = }\n")
print(f"{cov_40_36 = }")
print(f"{pho_40_36 = }\n")
print(f"{cov_39_36 = }")
print(f"{pho_39_36 = }\n")


ar36m = [float(line.rstrip("\n")) for line in open("ar36m.txt").readlines()]
ar37m = [float(line.rstrip("\n")) for line in open("ar37m.txt").readlines()]
ar38m = [float(line.rstrip("\n")) for line in open("ar38m.txt").readlines()]
ar39m = [float(line.rstrip("\n")) for line in open("ar39m.txt").readlines()]
ar40m = [float(line.rstrip("\n")) for line in open("ar40m.txt").readlines()]

cov_40m_39m = np.cov(ar40m, ar39m)
cov_40m_36m = np.cov(ar40m, ar36m)
cov_39m_36m = np.cov(ar39m, ar36m)

pho_40m_39m = cov_40m_39m[0][1] / (np.std(ar40m) * np.std(ar39m))
pho_40m_36m = cov_40m_36m[0][1] / (np.std(ar40m) * np.std(ar36m))
pho_39m_36m = cov_39m_36m[0][1] / (np.std(ar39m) * np.std(ar36m))

print(f"{cov_40m_39m = }")
print(f"{pho_40m_39m = }\n")
print(f"{cov_40m_36m = }")
print(f"{pho_40m_36m = }\n")
print(f"{cov_39m_36m = }")
print(f"{pho_39m_36m = }\n")


ar36a = [float(line.rstrip("\n")) for line in open("ar36a.txt").readlines()]
ar36cl = [float(line.rstrip("\n")) for line in open("ar36cl.txt").readlines()]
ar37ca = [float(line.rstrip("\n")) for line in open("ar37ca.txt").readlines()]
ar39k = [float(line.rstrip("\n")) for line in open("ar39k.txt").readlines()]
ar40r = [float(line.rstrip("\n")) for line in open("ar40r.txt").readlines()]

cov_40r_39k = np.cov(ar40r, ar39k)
cov_40r_36a = np.cov(ar40r, ar36a)
cov_39k_36a = np.cov(ar39k, ar36a)

pho_40r_39k = cov_40r_39k[0][1] / (np.std(ar40r) * np.std(ar39k))
pho_40r_36a = cov_40r_36a[0][1] / (np.std(ar40r) * np.std(ar36a))
pho_39k_36a = cov_39k_36a[0][1] / (np.std(ar39k) * np.std(ar36a))


print(f"{cov_40r_39k = }")
print(f"{pho_40r_39k = }\n")
print(f"{cov_40r_36a = }")
print(f"{pho_40r_36a = }\n")
print(f"{cov_39k_36a = }")
print(f"{pho_39k_36a = }\n")

"""
cov = array([[98.75793884, -0.31788865],
       [-0.31788865,  0.25810407]])
pho = -0.06297019225126861

误差相关性为负数

"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @File   : sample.py
# @Author : Yang Wu
# @Date   : 2021/12/19
# @Email  : wuy@cug.edu.cn

"""
# ==========================================
# Copyright 2023 Yang
# ararpy - smp - sample
# ==========================================
#
#
#

Create a sample instance.

"""
from typing import List, Tuple, Dict, Optional, Union
from types import MethodType
import pandas as pd

SAMPLE_INTERCEPT_HEADERS = [
    'Sequence', '',  # 0-1
    '\u00B3\u2076Ar', '1\u03C3', '\u00B3\u2077Ar', '1\u03C3',  # 2-5
    '\u00B3\u2078Ar', '1\u03C3', '\u00B3\u2079Ar', '1\u03C3',  # 6-9
    '\u2074\u2070Ar', '1\u03C3',  # 10-11
]
BLANK_INTERCEPT_HEADERS = [
    'Sequence', '',  # 0-1
    '\u00B3\u2076Ar', '1\u03C3', '\u00B3\u2077Ar', '1\u03C3',  # 2-5
    '\u00B3\u2078Ar', '1\u03C3', '\u00B3\u2079Ar', '1\u03C3',  # 6-9
    '\u2074\u2070Ar', '1\u03C3',  # 10-11
]
CORRECTED_HEADERS = [
    'Sequence', '',  # 0-1
    '\u00B3\u2076Ar', '1\u03C3', '\u00B3\u2077Ar', '1\u03C3',  # 2-5
    '\u00B3\u2078Ar', '1\u03C3', '\u00B3\u2079Ar', '1\u03C3',  # 6-9
    '\u2074\u2070Ar', '1\u03C3',  # 10-11
]
DEGAS_HEADERS = [
    'Sequence', '',  # 0-1
    '\u00B3\u2076Ar[a]', '1\u03C3', '\u00B3\u2076Ar[c]', '1\u03C3',  # 2-5
    '\u00B3\u2076Ar[Ca]', '1\u03C3', '\u00B3\u2076Ar[Cl]', '1\u03C3',  # 6-9
    '\u00B3\u2077Ar[Ca]', '1\u03C3', '\u00B3\u2078Ar[Cl]', '1\u03C3',  # 10-13
    '\u00B3\u2078Ar[a]', '1\u03C3', '\u00B3\u2078Ar[c]', '1\u03C3',  # 14-17
    '\u00B3\u2078Ar[K]', '1\u03C3', '\u00B3\u2078Ar[Ca]', '1\u03C3',  # 18-21
    '\u00B3\u2079Ar[K]', '1\u03C3', '\u00B3\u2079Ar[Ca]', '1\u03C3',  # 22-25
    '\u2074\u2070Ar[r]', '1\u03C3', '\u2074\u2070Ar[a]', '1\u03C3',  # 26-29
    '\u2074\u2070Ar[c]', '1\u03C3', '\u2074\u2070Ar[K]', '1\u03C3'  # 30-33
]
PUBLISH_TABLE_HEADERS = [
    'Sequence', '',  # 0-1
    '\u00B3\u2076Ar[a]', '\u00B3\u2077Ar[Ca]', '\u00B3\u2078Ar[Cl]',  # 2-4
    '\u00B3\u2079Ar[K]', '\u2074\u2070Ar[r]',  # 5-6
    'Apparent Age', '1\u03C3', '\u2074\u2070Arr%', '\u00B3\u2079ArK%',  # 7-10
    'Ca/K', '1\u03C3',  # 11-12
]
SPECTRUM_TABLE_HEADERS = [
    'Sequence', '',  # 0-1
    '\u2074\u2070Ar/\u00B3\u2079Ar', '1\u03C3',  # 2-3
    'Apparent Age', '1\u03C3', '1\u03C3', '1\u03C3',  # 4-7
    '\u2074\u2070Ar[r]%', '\u00B3\u2079Ar[K]%',  # 8-9
]
ISOCHRON_TABLE_HEADERS = [
    'Sequence', '', 'Mark',  # 0-2
    '\u00B3\u2079Ar[K]/\u00B3\u2076Ar[a]', '1\u03C3',  # 3-4
    '\u2074\u2070Ar[r+a]/\u00B3\u2076Ar[a]', '1\u03C3', 'ri', '',  # 5-8
    '\u00B3\u2079Ar[K]/\u2074\u2070Ar[r+a]', '1\u03C3',  # 9-10
    '\u00B3\u2076Ar[a]/\u2074\u2070Ar[r+a]', '1\u03C3', 'ri', '',  # 11-14
    '\u00B3\u2079Ar[K]/\u00B3\u2078Ar[Cl]', '1\u03C3',  # 15-16
    '\u2074\u2070Ar[r+Cl]/\u00B3\u2078Ar[Cl]', '1\u03C3', 'ri', '',  # 17-20
    '\u00B3\u2079Ar[K]/\u2074\u2070Ar[r+Cl]', '1\u03C3',  # 21-22
    '\u00B3\u2078Ar[Cl]/\u2074\u2070Ar[r+Cl]', '1\u03C3', 'ri', '',  # 23-26
    '\u00B3\u2078Ar[Cl]/\u00B3\u2079Ar[K]', '1\u03C3',  # 27-28
    '\u2074\u2070Ar[r+Cl]/\u00B3\u2079Ar[K]', '1\u03C3', 'ri', '',  # 29-32
    '\u00B3\u2076Ar[a+Cl]/\u2074\u2070Ar[r+a]', '1\u03C3',  # 33-34
    '\u00B3\u2078Ar[a+Cl]/\u2074\u2070Ar[r+a]', '1\u03C3',  # 35-36
    '\u00B3\u2079Ar[K]/\u2074\u2070Ar[r+a]', '1\u03C3',  # 37-38
    'r1', 'r2', 'r3',  # 39-41
]
TOTAL_PARAMS_HEADERS = [
    'Sequence', '',  # 0-1
    '(\u2074\u2070Ar/\u00B3\u2076Ar)t', '%1\u03C3',
    '(\u2074\u2070Ar/\u00B3\u2076Ar)c', '%1\u03C3',  # 2-5
    '(\u00B3\u2078Ar/\u00B3\u2076Ar)t', '%1\u03C3',
    '(\u00B3\u2078Ar/\u00B3\u2076Ar)c', '%1\u03C3',  # 6-9
    '(\u00B3\u2079Ar/\u00B3\u2077Ar)Ca', '%1\u03C3',
    '(\u00B3\u2078Ar/\u00B3\u2077Ar)Ca', '%1\u03C3',
    '(\u00B3\u2076Ar/\u00B3\u2077Ar)Ca', '%1\u03C3',  # 10-15
    '(\u2074\u2070Ar/\u00B3\u2079Ar)K', '%1\u03C3',
    '(\u00B3\u2078Ar/\u00B3\u2079Ar)K', '%1\u03C3',  # 16-19
    '(\u00B3\u2076Ar/\u00B3\u2078Ar)Cl', '%1\u03C3',  # 20-21
    'K/Ca', '%1\u03C3', 'K/Cl', '%1\u03C3', 'Ca/Cl', '%1\u03C3',  # 22-27
    'Cycle Number', 'Irradiation Cycles',  # 28-29
    'Irradiation', 'duration', 'Irradiation Time', 'Experiment Time',  # 30-33
    'Storage Years', '',  # 34-35
    'Decay Constant \u2074\u2070K', '%1\u03C3',  # 36-37
    'Decay Constant \u2074\u2070K(EC)', '%1\u03C3',  # 38-39
    'Decay Constant \u2074\u2070K(\u03B2<sup>-</sup>)', '%1\u03C3',  # 40-41
    'Decay Constant \u2074\u2070K(\u03B2<sup>+</sup>)', '%1\u03C3',  # 42-43
    'Decay Constant \u00B3\u2079Ar', '%1\u03C3',  # 44-45
    'Decay Constant \u00B3\u2077Ar', '%1\u03C3',  # 46-47
    'Decay Constant \u00B3\u2076Cl', '%1\u03C3',  # 48-49
    'Decay Activity \u2074\u2070K', '%1\u03C3',  # 50-51
    'Decay Activity \u2074\u2070K(EC)', '%1\u03C3',  # 52-53
    'Decay Activity \u2074\u2070K(\u03B2<sup>-</sup>)', '%1\u03C3',  # 54-55
    'Decay Activity \u2074\u2070K(\u03B2<sup>+</sup>)', '%1\u03C3',  # 56-57
    '\u00B3\u2076Cl/\u00B3\u2078Cl Productivity', '%1\u03C3',  # 58-59
    'Std Name', 'Std Age', '1\u03C3', '\u2074\u2070Ar%', '1\u03C3', 'K%', '1\u03C3',
    '\u2074\u2070Ar<sup>*</sup>/K', '1\u03C3',  # 60-68
    'J', '%1\u03C3', 'MDF', '%1\u03C3',  # 69-72
    'Mass \u00B3\u2076Ar', '%1\u03C3', 'Mass \u00B3\u2077Ar', '%1\u03C3',
    'Mass \u00B3\u2078Ar', '%1\u03C3', 'Mass \u00B3\u2079Ar', '%1\u03C3',
    'Mass \u2074\u2070', '%1\u03C3', 'K Mass', '%1\u03C3',  # 73-84
    'No', '%1\u03C3', 'Year', '%1\u03C3', '\u2074\u2070K/K', '%1\u03C3',
    '\u00B3\u2075/\u00B3\u2077Cl', '%1\u03C3', 'HCl/Cl', '%1\u03C3',  # 85-94
    '\u2074\u2070Ar/\u00B3\u2076Ar air', '%1\u03C3',
    '\u00B3\u2078Ar/\u00B3\u2076Ar air', '%1\u03C3',  # 95-98
    'Isochron Fitting', 'Convergence', 'Iteration', 'Discrimination',  # 99-102
    'Not Zero', 'Corr Blank', 'Corr Discr', 'Corr \u00B3\u2077Ar Decay',  # 103-106
    'Corr \u00B3\u2079Ar Decay',  # 107
    'Ca Degassing', 'K Degassing', 'Cl Degassing', 'Trap Degassing',  # 108-111
    'Using Min Equation',  # 112
    # 'Recalibration', 'Using Std Age', 'Use Std Ratio',  # 112-115  to be completed
    'Apply Gain Corr to Blanks',  # 113
    '', '',  # 114-115
    'Auto Plateau Method',  # 116 the index includes sequence name and unit
    'Initial Ratio Model',  # 117
    'Set1 initial Ratio',  # 118
    '1\u03C3',  # 119
    'Set2 initial Ratio',  # 120
    '1\u03C3',  # 121
    'Isotopic Errors',  # 122
    'Parameter Errors',  # 123
    'Plot Errors',  # 124
    'Heating Time (s)',  # 125
    'Heating Actual Temp (C)',  # 126
    'Heating AT 1\u03C3',  # 127
    '36Ar Gain', '%1\u03C3',  # 128-129
    '37Ar Gain', '%1\u03C3',  # 130-131
    '38Ar Gain', '%1\u03C3',  # 132-133
    '39Ar Gain', '%1\u03C3',  # 134-135
    '40Ar Gain', '%1\u03C3',  # 136-137
]

SAMPLE_INTERCEPT_SHORT_HEADERS = [
    'Seq', 'Label', 'Ar36', '1s', 'Ar37', '1s',
    'Ar38', '1s', 'Ar39', '1s', 'Ar40', '1s'
]
BLANK_INTERCEPT_SHORT_HEADERS = [
    'Seq', 'Label', 'Ar36', '1s', 'Ar37', '1s',
    'Ar38', '1s', 'Ar39', '1s', 'Ar40', '1s'
]
CORRECTED_SHORT_HEADERS = [
    'Seq', 'Label', 'Ar36', '1s', 'Ar37', '1s',
    'Ar38', '1s', 'Ar39', '1s', 'Ar40', '1s'
]
DEGAS_SHORT_HEADERS = [
    'Seq', 'Label',  # 0-1
    'Ar36[a]', '1s', 'Ar36[c]', '1s',  # 2-5
    'Ar36[Ca]', '1s', 'Ar36[Cl]', '1s',  # 6-9
    'Ar37[Ca]', '1s', 'Ar38[Cl]', '1s',  # 10-13
    'Ar38[a]', '1s', 'Ar38[c]', '1s',  # 14-17
    'Ar38[K]', '1s', 'Ar38[Ca]', '1s',  # 18-21
    'Ar39[K]', '1s', 'Ar39[Ca]', '1s',  # 22-25
    'Ar40[r]', '1s', 'Ar40[a]', '1s',  # 26-29
    'Ar40[c]', '1s', 'Ar40[K]', '1s'  # 29-32
]
PUBLISH_TABLE_SHORT_HEADERS = [
    'Seq', 'Label', 'Ar36[a]', 'Ar37[Ca]', 'Ar38[Cl]',
    'Ar39[K]', 'Ar40[r]', 'Apparent Age', '1s', 'Ar40r%',
    'Ar39K%', 'Ca/K', '1s'
]
SPECTRUM_TABLE_SHORT_HEADERS = [
    'Seq', 'Label', 'Ar40/Ar39', '1s', 'Apparent Age',
    '1s', '1s', '1s', 'Ar40[r]%', 'Ar39[K]%'
]
ISOCHRON_TABLE_SHORT_HEADERS = [
    'Seq', 'Label', 'Mark',  # 0-2
    'Ar39[K]/Ar36[a]', '1s',  # 3-4
    'Ar40[r+a]/Ar36[a]', '1s', 'ri', '',  # 5-8
    'Ar39[K]/Ar40[r+a]', '1s',  # 9-10
    'Ar36[a]/Ar40[r+a]', '1s', 'ri', '',  # 11-14
    'Ar39[K]/Ar38[Cl]', '1s',  # 15-16
    'Ar40[r+Cl]/Ar38[Cl]', '1s', 'ri', '',  # 17-20
    'Ar39[K]/Ar40[r+Cl]', '1s',  # 21-22
    'Ar38[Cl]/Ar40[r+Cl]', '1s', 'ri', '',  # 23-26
    'Ar38[Cl]/Ar39[K]', '1s',  # 27-28
    'Ar40[r+Cl]/Ar39[K]', '1s', 'ri', '',  # 29-32
    'Ar36[a+Cl]/Ar40[r+a]', '1s',  # 33-34
    'Ar38[a+Cl]/Ar40[r+a]', '1s',  # 35-36
    'Ar39[K]/Ar40[r+a]', '1s',  # 37-38
    'r1', 'r2', 'r3',  # 39-41
]
TOTAL_PARAMS_SHORT_HEADERS = [
    'Seq', 'Label',  # 0-1
    'T_40v36', '%1s',
    'C_40v36', '%1s',  # 2-5
    'T_38v36', '%1s',
    'C_38v36', '%1s',  # 6-9
    'Ca_39v37', '%1s',
    'Ca_38v37', '%1s',
    'Ca_36v37', '%1s',  # 10-15
    'K_40v39', '%1s',
    'K_38v39', '%1s',  # 16-19
    'Cl_36v38', '%1s',  # 20-21
    'KvCa', '%1s', 'KvCl', '%1s', 'CavCl', '%1s',  # 22-27
    'Ncy', 'Icy',  # 28-29
    'Irr', 'Dur', 'IrrTime', 'ExpTime',  # 30-33
    'StgY', '',  # 34-35
    'DCK40', '%1s',  # 36-37
    'DCeK40', '%1s',  # 38-39
    'DCb-K40', '%1s',  # 40-41
    'DCb+K40', '%1s',  # 42-43
    'DCAr39', '%1s',  # 44-45
    'DCAr37', '%1s',  # 46-47
    'DCCl36', '%1s',  # 48-49
    'DAK40', '%1s',  # 50-51
    'DAeK40', '%1s',  # 52-53
    'DAb-K40', '%1s',  # 54-55
    'DAb+K40', '%1s',  # 56-57
    'Cl36/Cl38P', '%1s',  # 58-59
    'StdName', 'StdAge', '1s', 'Ar40%', '1s', 'K%', '1s',
    'F', '1s',  # 60-68
    'J', '%1s', 'MDF', '%1s',  # 69-72
    'MassAr36', '%1s', 'MassAr37', '%1s',
    'MassAr38', '%1s', 'MassAr39', '%1s',
    'MassK40', '%1s', 'MassK', '%1s',  # 73-84
    'No', '%1s', 'Year', '%1s', 'K40/K', '%1s',
    'Cl35/Cl37', '%1s', 'HCl/Cl', '%1s',  # 85-94
    'Ar40/Ar36Air', '%1s',
    'Ar38/Ar36Air', '%1s',  # 95-98
    'IsochronFitting', 'Convergence', 'Iteration', 'Discrimination',  # 99-102
    'NotZero', 'CorrBlank', 'CorrDiscr', 'CorrAr37Decay',
    'CorrAr39Decay',  # 103-107
    'CaDegassing', 'KDegassing', 'ClDegassing', 'TrapDegassing',  # 108-111
    'UsingMin',  # 112
    # 'Recalibration', 'Using Std Age', 'Use Std Ratio',  # 112-115  to be completed
    'BlankGainCorr',  # 113
    '', '',  # 114-115
    'AutoPlateauMethod',  # 116 the index includes sequence name and unit
    'InitialRatioModel',  # 117
    'Set1InitialRatio',  # 118
    '1s',  # 119
    'Set2InitialRatio',  # 120
    '1s',  # 121
    'IsotopicErrors',  # 122
    'ParameterErrors',  # 123
    'PlotErrors',  # 124
    'HeatingTime',  # 125
    'HeatingActualTemp',  # 126
    'HeatingActualTempError',  # 127
    '36Gain', '%1s',  # 128-129
    '37Gain', '%1s',  # 130-131
    '38Gain', '%1s',  # 132-133
    '39Gain', '%1s',  # 134-135
    '40Gain', '%1s',  # 136-137
]

DEFAULT_PLOT_STYLES = {
    'figure_1': {
        'id': 'figure_1', 'name': 'Age Spectra', 'type': 'spectra', 'attr_name': 'AgeSpectraPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': 'Age Spectra', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text', 'font_weight': 'bolder',
                  'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': 'Cumulative {sup|39}Ar Released (%)', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': 'Apparent Age (Ma)', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'set1': {'id': 'Points Set 1', 'type': 'set', },
        'set2': {'id': 'Points Set 2', 'type': 'set', },
        'set3': {'id': 'Points Set 3', 'type': 'set', },
        'set4': {'id': 'Points Set 4', 'type': 'set', },
        'set5': {'id': 'Points Set 5', 'type': 'set', },
        'line1': {
            'id': 'Spectra Line 1', 'color': '#333333', 'line_type': 'solid',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line2': {
            'id': 'Spectra Line 2', 'color': '#333333', 'line_type': 'solid',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line3': {
            'id': 'Set1 Line 1', 'color': '#FF0000', 'line_type': 'solid',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line4': {
            'id': 'Set1 Line 2', 'color': '#FF0000', 'line_type': 'solid',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line5': {
            'id': 'Set2 Line 1', 'color': '#00B0F0', 'line_type': 'solid',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line6': {
            'id': 'Set2 Line 2', 'color': '#00B0F0', 'line_type': 'solid',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line7': {
            'id': 'Set4 Line 1', 'color': '#FF0000', 'line_type': 'dashed',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line8': {
            'id': 'Set4 Line 2', 'color': '#FF0000', 'line_type': 'dashed',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line9': {
            'id': 'Set5 Line 1', 'color': '#00B0F0', 'line_type': 'dashed',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'line10': {
            'id': 'Set5 Line 2', 'color': '#00B0F0', 'line_type': 'dashed',
            'line_width': 2, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                      'color': '#252525'}
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'text2': {
            'id': 'Text for Set 2', 'show': True, 'font_size': 16, 'color': '#00B0F0', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_2': {
        'id': 'figure_2', 'name': 'Normal Isochron', 'type': 'isochron', 'attr_name': 'NorIsochronPlot',
        'rightside_text': [],
        'ellipse': {
            'id': 'ellipse', 'color': 'none', 'border_color': '#333333', 'border_type': 'solid',
            'show': False, 'fill': False,
            'border_width': 1, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
        },
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': '{sup|39}Ar{sub|K} / {sup|36}Ar{sub|a}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': '{sup|40}Ar{sup|*} / {sup|36}Ar{sub|a}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'set1': {
            'id': 'Points Set 1', 'color': '#FF0000', 'border_color': '#FF0000', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set2': {
            'id': 'Points Set 2', 'color': '#00B0F0', 'border_color': '#00B0F0', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set3': {
            'id': 'Unselected Points', 'color': '#FFFFFF', 'border_color': '#222222', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line1': {
            'id': 'Line for Set 1', 'type': 'set', 'color': '#FF0000', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line2': {
            'id': 'Line for Set 2', 'type': 'set', 'color': '#00B0F0', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'errline': {
            'id': 'Error Lines', 'type': 'set', 'color': '#333333', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label'}, 'show': True
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'text2': {
            'id': 'Text for Set 2', 'show': True, 'font_size': 16, 'color': '#00B0F0', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_3': {
        'id': 'figure_3', 'name': 'Inverse Isochron', 'type': 'isochron', 'attr_name': 'InvIsochronPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': '{sup|39}Ar{sub|K} / {sup|40}Ar{sup|*}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': '{sup|36}Ar{sub|a} / {sup|40}Ar{sup|*}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'ellipse': {
            'id': 'ellipse', 'color': 'none', 'border_color': '#333333', 'border_type': 'solid',
            'show': False, 'fill': False,
            'border_width': 1, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
        },
        'set1': {
            'id': 'Points Set 1', 'color': '#FF0000', 'border_color': '#FF0000', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set2': {
            'id': 'Points Set 2', 'color': '#00B0F0', 'border_color': '#00B0F0', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set3': {
            'id': 'Unselected Points', 'color': '#FFFFFF', 'border_color': '#222222', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line1': {
            'id': 'Line for Set 1', 'type': 'set', 'color': '#FF0000', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label', }
        },
        'line2': {
            'id': 'Line for Set 2', 'type': 'set', 'color': '#00B0F0', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label', }
        },
        'errline': {
            'id': 'Error Lines', 'type': 'set', 'color': '#333333', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label'}, 'show': True
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'text2': {
            'id': 'Text for Set 2', 'show': True, 'font_size': 16, 'color': '#00B0F0', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_4': {
        'id': 'figure_4', 'name': 'K-Cl-Ar 1', 'type': 'isochron', 'attr_name': 'KClAr1IsochronPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': '{sup|39}Ar{sub|K} / {sup|38}Ar{sub|Cl}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': '{sup|40}Ar{sup|*} / {sup|38}Ar{sub|Cl}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'ellipse': {
            'id': 'ellipse', 'color': 'none', 'border_color': '#333333', 'border_type': 'solid',
            'show': False, 'fill': False,
            'border_width': 1, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
        },
        'set1': {
            'id': 'Points Set 1', 'color': '#FF0000', 'border_color': '#FF0000', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set2': {
            'id': 'Points Set 2', 'color': '#00B0F0', 'border_color': '#00B0F0', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set3': {
            'id': 'Unselected Points', 'color': '#FFFFFF', 'border_color': '#222222', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line1': {
            'id': 'Line for Set 1', 'type': 'set', 'color': '#FF0000', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line2': {
            'id': 'Line for Set 2', 'type': 'set', 'color': '#00B0F0', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'errline': {
            'id': 'Error Lines', 'type': 'set', 'color': '#333333', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label'}, 'show': True
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'text2': {
            'id': 'Text for Set 2', 'show': True, 'font_size': 16, 'color': '#00B0F0', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_5': {
        'id': 'figure_5', 'name': 'K-Cl-Ar 2', 'type': 'isochron', 'attr_name': 'KClAr2IsochronPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': '{sup|39}Ar{sub|K} / {sup|40}Ar{sup|*}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': '{sup|38}Ar{sub|Cl} / {sup|40}Ar{sup|*}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'ellipse': {
            'id': 'ellipse', 'color': 'none', 'border_color': '#333333', 'border_type': 'solid',
            'show': False, 'fill': False,
            'border_width': 1, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
        },
        'set1': {
            'id': 'Points Set 1', 'color': '#FF0000', 'border_color': '#FF0000', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set2': {
            'id': 'Points Set 2', 'color': '#00B0F0', 'border_color': '#00B0F0', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set3': {
            'id': 'Unselected Points', 'color': '#FFFFFF', 'border_color': '#222222', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line1': {
            'id': 'Line for Set 1', 'type': 'set', 'color': '#FF0000', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line2': {
            'id': 'Line for Set 2', 'type': 'set', 'color': '#00B0F0', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'errline': {
            'id': 'Error Lines', 'type': 'set', 'color': '#333333', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label'}, 'show': True
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'text2': {
            'id': 'Text for Set 2', 'show': True, 'font_size': 16, 'color': '#00B0F0', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_6': {
        'id': 'figure_6', 'name': 'K-Cl-Ar 3', 'type': 'isochron', 'attr_name': 'KClAr3IsochronPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': '{sup|38}Ar{sub|Cl} / {sup|39}Ar{sub|K}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': '{sup|40}Ar{sup|*} / {sup|39}Ar{sub|K}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'ellipse': {
            'id': 'ellipse', 'color': 'none', 'border_color': '#333333', 'border_type': 'solid',
            'show': False, 'fill': False,
            'border_width': 1, 'opacity': 1, 'symbol_size': 0, 'symbol': 'none', 'type': 'set',
        },
        'set1': {
            'id': 'Points Set 1', 'color': '#FF0000', 'border_color': '#FF0000', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set2': {
            'id': 'Points Set 2', 'color': '#00B0F0', 'border_color': '#00B0F0', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set3': {
            'id': 'Unselected Points', 'color': '#FFFFFF', 'border_color': '#222222', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'line1': {
            'id': 'Line for Set 1', 'type': 'set', 'color': '#FF0000', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label'}
        },
        'line2': {
            'id': 'Line for Set 2', 'type': 'set', 'color': '#00B0F0', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label'}
        },
        'errline': {
            'id': 'Error Lines', 'type': 'set', 'color': '#333333', 'line_type': 'solid', 'line_width': 2,
            'label': {'show': False, 'type': 'label'}, 'show': True
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'text2': {
            'id': 'Text for Set 2', 'show': True, 'font_size': 16, 'color': '#00B0F0', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_7': {
        'id': 'figure_7', 'name': '3D Isochron', 'type': 'isochron', 'attr_name': 'ThreeDIsochronPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': '{sup|36}Ar{sub|a+Cl} / {sup|40}Ar{sub|a+r}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'zaxis': {
            'title': {'text': '{sup|39}Ar{sub|K} / {sup|40}Ar{sub|a+r}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': '{sup|38}Ar{sub|a+Cl} / {sup|40}Ar{sub|a+r}', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'set1': {
            'id': 'Points Set 1', 'color': '#FF0000', 'border_color': '#FF0000', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set2': {
            'id': 'Points Set 2', 'color': '#00B0F0', 'border_color': '#00B0F0', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'set3': {
            'id': 'Unselected Points', 'color': '#FFFFFF', 'border_color': '#222222', 'border_width': 2, 'type': 'set',
            'border_type': 'solid', 'opacity': 0.8, 'symbol_size': 10, 'symbol': 'circle',
            'label': {'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top', 'type': 'label'}
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'text2': {
            'id': 'Text for Set 2', 'show': True, 'font_size': 16, 'color': '#00B0F0', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif',
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_8': {
        'id': 'figure_8', 'name': 'Degas Pattern', 'type': 'plot', 'attr_name': 'DegasPatternPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': 'Step', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': 'Argon isotopes (%)', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                  'color': '#252525'}
    },
    'figure_9': {
        'id': 'figure_9', 'name': 'Age Distribution', 'type': 'plot', 'attr_name': 'AgeDistributionPlot',
        'rightside_text': [],
        'title': {'id': 'title', 'show': True, 'text': '', 'position': 'center',
                  'font_size': 18, 'color': '#333333', 'opacity': 1, 'type': 'text',
                  'font_weight': 'bolder', 'font_family': 'Microsoft Sans Serif', },
        'xaxis': {
            'title': {'text': '{sup|40}Ar / {sup|39}Ar Age', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'yaxis': {
            'title': {'text': 'Count', 'type': 'text', }, 'type': 'axis',
            'min': 0, 'max': 100, 'show_splitline': False, 'ticks_inside': False, 'split_number': 5, 'interval': 10,
        },
        'set1': {
            'id': 'Histogram', 'color': '#FFFFFF', 'border_color': '#333333', 'border_width': 1,
            'border_type': 'dashed', 'opacity': 0.5, 'symbol_size': 10, 'type': 'set', 'index': 0,
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0],
                      'position': 'top', 'color': '#252525'}
        },
        'set2': {
            'id': 'KDE', 'color': '#FF0000', 'line_width': 2, 'line_type': 'solid', 'symbol': 'None', 'symbol_size': 0,
            'opacity': 1, 'type': 'set', 'index': 9, 'band_extend': False, 'band_maximize': False, 'band_points': 200,
            'auto_width': 'Scott',
            'label': {'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0],
                      'position': 'top', 'color': '#252525'}
        },
        'set3': {
            'id': 'Age Bar', 'color': '#252525', 'border_color': '#252525', 'border_width': 1, 'border_type': 'solid',
            'opacity': 1, 'symbol_size': 10, 'type': 'set', 'index': 5, 'vertical_align': 'spread', 'bar_height': 5,
            'bar_interval': 5,
            'label': {
                'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
                'color': '#252525',
            }
        },
        'text1': {
            'id': 'Text for Set 1', 'show': True, 'font_size': 16, 'color': '#FF0000', 'opacity': 1, 'type': 'text',
            'font_weight': 'bold', 'font_family': 'Microsoft Sans Serif', 'pos': [1500, 10],
        },
        'label': {
            'id': '', 'type': 'label', 'show': False, 'distance': 5, 'offset': [10, 0], 'position': 'top',
            'color': '#252525',
        }
    },
}

VERSION = '20250328'

NAMED_DICT = {
    "unknown": {"header": SAMPLE_INTERCEPT_HEADERS.copy()},
    "blank": {"header": BLANK_INTERCEPT_HEADERS.copy()},
    "parameters": {"header": TOTAL_PARAMS_HEADERS.copy()},
    "corrected": {"header": CORRECTED_HEADERS.copy()},
    "degas": {"header": DEGAS_HEADERS.copy()},
    "isochron": {"header": ISOCHRON_TABLE_HEADERS.copy()},
    "apparent_ages": {"header": SPECTRUM_TABLE_HEADERS.copy()},
    "publish": {"header": PUBLISH_TABLE_HEADERS.copy()},
}


class ArArBasic:
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)


class ArArData(ArArBasic):
    def __init__(self, **kwargs):
        self.name: str = ""
        self.data: list = []
        self.header: list = []
        self.short_header: list = []
        super().__init__(**kwargs)
        if not isinstance(self.data, list):
            raise TypeError(f"Data must be a list.")
        if self.name.lower() in NAMED_DICT.keys() and self.header == []:
            self.header = NAMED_DICT[self.name]['header']
        if len(self.header) != len(self.data):
            self.header = [*self.header, *list(range(len(self.header), len(self.data)))][:len(self.data)]

    def to_df(self) -> pd.DataFrame: ...
    def to_list(self) -> list: ...


class Sample:

    def __init__(self, **kwargs):
        self.Doi = ""
        self.RawData = RawData()
        self.Info = ArArBasic()

        self.SequenceName = []
        self.SequenceValue = []
        self.SequenceUnit = []

        self.SampleIntercept = []
        self.BlankIntercept = []
        self.AnalysisDateTime = []

        self.BlankCorrected = []
        self.MassDiscrCorrected = []
        self.DecayCorrected = []
        self.CorrectedValues = []
        self.DegasValues = []

        self.ApparentAgeValues = []
        self.IsochronValues = []
        self.TotalParam = []

        self.PublishValues = []

        self.SelectedSequence1 = []
        self.SelectedSequence2 = []
        self.UnselectedSequence = []
        self.IsochronMark = []

        # Tables and Plots
        self.UnknownTable = Table()
        self.BlankTable = Table()
        self.CorrectedTable = Table()
        self.DegasPatternTable = Table()
        self.PublishTable = Table()
        self.AgeSpectraTable = Table()
        self.IsochronsTable = Table()
        self.TotalParamsTable = Table()
        self.AgeSpectraPlot = Plot()
        self.NorIsochronPlot = Plot()
        self.InvIsochronPlot = Plot()
        self.KClAr1IsochronPlot = Plot()
        self.KClAr2IsochronPlot = Plot()
        self.KClAr3IsochronPlot = Plot()
        self.ThreeDIsochronPlot = Plot()
        self.CorrelationPlot = Plot()
        self.DegasPatternPlot = Plot()
        self.AgeDistributionPlot = Plot()

        for k, v in kwargs.items():
            setattr(self, k, v)

        # self.Components = {}

        # self.__version = '20230521'
        # self.__version = '20230709'  # add labels to isochron plots
        # self.__version = '20230724'  # change header
        # self.__version = '20230730'  # delete calcparams attribute
        # self.__version = '20230827'  # using merge smp to update arr version
        # self.__version = '20231116'  # change smp parameters
        # self.__version = '20240730'  # change parameter table for thermo calculation
        # self.__version = '20241028'  # gain correction
        # self.__version = '20250102'  # gain correction to blanks
        # self.__version = '20250301'  # update sample info
        # self.__version = '20250321'  # error sigma adjustment
        self.__version = '20250328'  # Experiment info

    @property
    def version(self):
        return self.__version

    def help(self) -> str: ...
    
    def name(self) -> str: ...

    def doi(self) -> str: ...

    def sample(self) -> ArArBasic: ...

    def researcher(self) -> ArArBasic: ...

    def laboratory(self) -> ArArBasic: ...

    def results(self) -> ArArBasic: ...

    def sequence(self) -> ArArBasic: ...

    def recalculate(self, *args, **kwargs): ...

    def plot_age_plateau(self): ...

    def plot_isochron(self): ...

    def plot_normal(self): ...

    def plot_inverse(self): ...

    def plot_cl_1(self): ...

    def plot_cl_2(self): ...

    def plot_cl_3(self): ...

    def plot_3D(self): ...

    def initial(self): ...

    def set_selection(self): ...

    def update_table(self, data: list, table_id: name): ...

    def unknown(self) -> ArArData: ...

    def blank(self) -> ArArData: ...

    def parameters(self) -> ArArData: ...

    def corrected(self) -> ArArData: ...

    def degas(self) -> ArArData: ...

    def isochron(self) -> ArArData: ...

    def apparent_ages(self) -> ArArData: ...

    def publish(self) -> ArArData: ...

    def corr_blank(self): ...

    def corr_massdiscr(self): ...

    def corr_decay(self): ...

    def corr_ca(self): ...

    def corr_k(self): ...

    def corr_cl(self): ...

    def corr_atm(self): ...

    def corr_r(self): ...

    def calc_ratio(self): ...

    def show_data(self): ...

    def set_params(self, params: Union[List, str], flag: Optional[str] = None, rows: Optional[List] = None): ...

    def set_info(self, info: dict): ...

    def get_comp(self, name: str):...

    def to_pdf(self, file_path: str, figure: str = "figure_3"): ...

    def to_excel(self, file_path: str, *args, **kwargs): ...


class Table:
    def __init__(self, id='', name='Table', colcount=None, rowcount=None, header=None,
                 data=None, coltypes=None, text_indexes=None, numeric_indexes=None, **kwargs):
        if header is None:
            header = ['']
        if data is None:
            data = [['' for i in range(len(header))]]
        if colcount is None:
            colcount = len(header)
        if rowcount is None:
            rowcount = len(data[0])
        if coltypes is None:
            # Note the difference between [xx] * 10 and [xx for i in range(10)]
            coltypes = [{'type': 'text'} for i in range(colcount)]
        self.id = id
        self.name = name
        self.colcount = colcount
        self.rowcount = rowcount
        self.header = header
        self.data = data
        self.coltypes = coltypes
        self.text_indexes = text_indexes if text_indexes is not None else []
        self.numeric_indexes = numeric_indexes if numeric_indexes is not None else list(set(range(0, colcount)) - set(self.text_indexes))
        self.decimal_places = 6

        for k, v in kwargs.items():
            setattr(self, k, v)
        self.set_coltypes()

    def set_coltypes(self):
        for i in self.text_indexes:
            if i < self.colcount:
                self.coltypes[i].update({'type': 'text'})
        for i in self.numeric_indexes:
            if i < self.colcount:
                self.coltypes[i].update({
                    'type': 'numeric',
                    'numericFormat': {'pattern': {'mantissa': self.decimal_places}},
                })


class Plot:
    def __init__(self, id='', type='', name='', data=None, info=None, **kwargs):
        if data is None:
            data = []
        self.id = id
        self.data = data
        self.info = info
        self.type = type
        self.name = name
        for k, v in kwargs.items():
            setattr(self, k, v)

    class BasicAttr:
        def __init__(self, enabled=True, id='', type='', info=None, ):
            self.enabled = enabled
            self.id = id
            self.type = type
            self.info = info

        def show(self):
            self.enabled = True
            return self

        def hidden(self):
            self.enabled = False
            return self

    class Axis(BasicAttr):
        def __init__(self, id='', title=None, min=0, max=100, show_splitline=False,
                     **kwargs):
            super().__init__(id=id, type='Axis')
            self.title = title
            self.min = min
            self.max = max
            self.show_splitline = show_splitline
            for k, v in kwargs.items():
                setattr(self, k, v)

    class Text(BasicAttr):
        def __init__(self, id='', text='', show=True, color=None, font_size=8,
                     font_family=None, font_weight=None,  # bold, normal
                     pos=None, **kwargs):
            super().__init__(id=id, type='Text')
            if pos is None:
                pos = [0, 0]
            self.text = text
            self.show = show
            self.color = color
            self.font_family = font_family
            self.font_size = font_size
            self.font_weight = font_weight
            self.pos = pos
            for k, v in kwargs.items():
                setattr(self, k, v)

    class Label(BasicAttr):
        def __init__(self, id='', show=False, color=None, position=None, distance=None,
                     offset=None, **kwargs):
            super().__init__(id=id, type='Label')
            if offset is None:
                offset = [10, 0]
            self.show = show
            self.color = color
            self.position = position
            self.offset = offset
            self.distance = distance
            for k, v in kwargs.items():
                setattr(self, k, v)

    class Set(BasicAttr):
        def __init__(self, id='', label=None, info=None, color=None, border_color=None,
                     border_width=None, line_width=None, line_type=None, opacity=None,
                     index=None, data=None, symbol_size=None, **kwargs):
            super().__init__(id=id, type='Set', info=info)
            if data is None:
                data = []
            self.label = label
            self.color = color
            self.border_color = border_color
            self.border_width = border_width
            self.line_width = line_width
            self.line_type = line_type
            self.opacity = opacity
            self.index = index
            self.data = data
            self.symbol_size = symbol_size
            for k, v in kwargs.items():
                setattr(self, k, v)


class Sequence:
    def __init__(self, index=None, data=None, flag=None, name=None, datetime=None,
                 type_str=None, results=None, coefficients=None, fitting_method=None,
                 is_estimated=False, options=None, **kwargs):
        if options is None:
            options = {}
        self.index = index
        if name is None or not isinstance(name, str):
            name = ""
        self.name = name.strip()
        self.datetime = datetime
        self.data = data
        # flag is to check if the data point is selected
        if flag is None and data is not None:
            flag = [[j if i == 0 else True for i, j in enumerate(_m)] for _m in data]
        self.flag = flag
        self.type_str = type_str
        if results is None:
            results = []
        if fitting_method is None:
            fitting_method = []
        if coefficients is None:
            coefficients = []
        self.results = results
        self.coefficients = coefficients
        self.fitting_method = fitting_method
        self.is_estimated = is_estimated
        self.is_removed = False
        self.options = options

        for k, v in kwargs.items():
            if hasattr(self, k) and type(getattr(self, k)) is MethodType:
                continue
            setattr(self, k, v)

        self.__as_type(self.type_str)

    def as_type(self, type_str):
        if str(type_str).lower() in ["blk", "b", "blank"]:
            self.type_str = "blank"
        if str(type_str).lower() in ["a", "air"]:
            self.type_str = "air"
        if self.type_str not in ["blank", "air"]:
            self.type_str = "unknown"

    __as_type = as_type

    def is_blank(self):
        return self.type_str == "blank"

    def is_unknown(self):
        return self.type_str != "blank" and self.type_str != "air"

    def is_air(self):
        return self.type_str == "air"

    def as_blank(self):
        return self.as_type("blank")

    def as_unknown(self):
        return self.as_type("unknown")

    def as_air(self):
        return self.as_type("air")

    def get_data_df(self):
        ...

    def get_flag_df(self):
        ...


class RawData:
    def __init__(self, id='', name='', type='Unknown', data=None, sequence_num=0,
                 isotopic_num=10, source=None, sequence=None, unit='fA', **kwargs):
        """
        Parameters
        ----------
        id
        name : file name
        type : raw
        data :
            list or data frame. [[x_ar36, y_ar36, x_ar37, y_ar37, ...], [ ... ]
        kwargs
        """

        self.id = id
        self.name = name
        self.type = type
        self.source = source
        self.unit = unit
        self.isotopic_num = isotopic_num
        self.sequence_num = sequence_num
        self.interpolated_blank = None
        if sequence is not None:
            self.sequence = sequence
        elif data is not None:
            self.sequence: List[Sequence] = [
                Sequence(
                    index=index,
                    name=item[0][0] if isinstance(item[0][0], str) and item[0][0] != '' else f"{self.name}-{index + 1:02d}",
                    data=item[1:],
                    datetime=item[0][1],
                    type_str=item[0][2],
                    fitting_method=[*kwargs.get("fitting_method", [0] * 5)],
                    options=item[0][3]
                ) for index, item in enumerate(data)]
        else:
            self.sequence: List[Sequence] = []
        for k, v in kwargs.items():
            if hasattr(self, k) and type(getattr(self, k)) is MethodType:
                continue

    def set_flag(self, sequence_index: int, isotopic_index: int, data: List[List[float]]):
        ...

    def get_flag(self, sequence_index: Optional[int], isotopic_index: Optional[int]) -> List[List[float]]:
        ...

    def get_result(self, sequence_index: Optional[int], isotopic_index: Optional[int],
                   method_index: Optional[List[List[Union[int, str]]]]) -> List[List[float]]:
        ...

    def do_regression(self, sequence_index: Optional[List], isotopic_index: Optional[List]):
        ...

    def get_data_df(self):
        ...

    def get_sequence(self, index: Optional[Union[list, int, str, bool]], flag: Optional[str],
                     unique: Optional[bool] = True) -> Union[Sequence, List]:
        ...

    def get_unknown(self) -> Union[Sequence, List]:
        ...

    def get_blank(self) -> Union[Sequence, List]:
        ...

    def get_air(self) -> Union[Sequence, List]:
        ...

    def to_sample(self, mapping: Optional[List[dict]]) -> Sample:
        ...



#  Copyright (C) 2024 Yang. - All Rights Reserved

#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2024 Yang 
# ararpy - test2
# ==========================================
#
#
# 
"""
import os
import ararpy as ap
import numpy as np
import scipy
from math import exp, log, sqrt

# kfcs = [1.634, 1.639, 1.665, 1.641, 1.639, 1.641]
# skfcs = [0.023, 0.016, 0.023, 0.010, 0.008, 0.009]
#
# res = ap.calc.arr.wtd_mean(kfcs, skfcs, sf=1, adjust_error=False)
# # res =  (1.6408482971738128, 0.004679971790869418, 6, 0.251709745316985, 1.258548726584925, 0.9391386247331938)
# print(res)

LE = 0.580E-10
sLE = 0.007E-10
LB = 4.884E-10
sLB = 0.049E-10
LT = 5.5545E-10
sLT = 0.0109E-10
K = 1.6408E-3
sK = 0.0047E-3

T = [
    0.001919,
    17.42, 125.65, 133.06, 201.27, 242.14, 252.17, 252.4,
    252.6, 253.02, 253.24, 253.68, 257.3, 327.99, 454.59, 1094.2, 2067.5
]

sT = [
    0.0000000001,
    0.03, 0.17, 0.26, 0.13, 0.45, 0.36, 0.33, 0.33, 0.5,
    0.33, 0.3, 0.3, 0.45, 0.56, 1.15, 1.4
]

T[1:] = [i - 0.09 for i in T[1:]]
sT[1:] = [sqrt(i ** 2 + 0.077 ** 2) for i in sT[1:]]

R = [
    6.789E-05,
    0.60556, 4.5494, 4.8205, 7.4636, 9.0619, 9.49831, 9.4918, 9.51018,
    9.4958, 9.52952, 9.53910, 9.69866, 12.60389, 18.09, 52.9011, 135.63893
]

sR = [
    5.50E-07,
    0.00249, 0.0061, 0.01535, 0.0072, 0.0122, 0.01022, 0.0041, 0.00594,
    0.0122, 0.00680, 0.00846, 0.00532, 0.01475, 0.0523, 0.1162, 0.18206
]

Opt_R = [
    6.73E-05, 0.6076, 4.5531, 4.8319, 7.4616, 9.0676, 9.4935, 9.4919, 9.5087, 9.5016, 9.5289, 9.5408, 9.6975, 12.604, 18.114, 52.796, 135.72
]

Opt_LE = 0.5755E-10
Opt_sLE = 0.0016E-10
Opt_LB = 4.9737E-10
Opt_sLB = 0.0093E-10
Opt_LT = 5.5492E-10
Opt_sLT = 0.0093E-10
Opt_K = 1.6418E-3
Opt_sK = 0.0045E-3
# R = 4.82050
# sR = 0.01535
# t = 133.06
# st = 0.26


def square(a, b, sb):
    return ((a - b) / sb) ** 2


def get_ti(_le, _lb, _k, _ri):
    return log((_le + _lb) * _k * _ri / _le + 1) / (_le + _lb) / 1000000


def get_standards(random: bool = False):


    return T, R, LE, LB, K, LT


def func(args):
    le, lb, k = args[:3]

    # r = R
    r = args[3:]
    # ti = args[3:int(3 + len(args[3:]) / 2)]
    # ri = args[int(3 + len(args[3:]) / 2):]

    t = [get_ti(le, lb, k, ri) for ri in r]

    return sum([square(ti, T[i], sT[i]) / 2 + square(r[i], R[i], sR[i]) / 2 for i, ti in enumerate(t)]
               ) + square(le, LE, sLE) + square(lb, LB, sLB) + square(k, K, sK) + square(le + lb, LT, sLT)

    # return sum([(square(ti[i], ti[i], st) + square(ri[i], ri[i], sR)) / 2 for i in range(len(ti))]
    #            ) + square(le, LE, sLE) + square(lb, LB, sLB) + square(k, K, sK) + square(le + lb, LT, sLT)

    # return (square(le, LE, sLE) + square(lb, LB, sLB) + square(k, K, sK) + square(le + lb, LT, sLT)) * 17


def MF_regerssion(x0):
    # print("\n ======== 最小值结果 ======== \n")
    a = scipy.optimize.minimize(fun=func, x0=x0, method="Nelder-Mead", options={"maxiter": 10000}, tol=0.00000000001)
    return a["x"].tolist()


def calc_age(R: float, sR: float, le: list, lb: list, k: list):
    R = np.random.normal(R, sR, len(le))
    res = [1 / (le[i] + lb[i]) * log((le[i] + lb[i]) / le[i] * k[i] * R[i] + 1) / 1000000 for i in range(len(le))]
    return np.mean(res), np.std(res)


def get_statistic(a0):
    k0 = np.mean(a0)
    k1 = np.std(a0)
    k2 = np.var(a0)

    return k0, k1, k2


def this_is_to_identify_that_monte_carlo_is_not_better_than_linear(R = 50, sR = 0.25):
    """
    Renne 2010 conclude that using their results people should using monte carlo method to estimate uncertainty in ages,
    however, it is a fact that their random values of LE, LB and K are generated according to the variances and covariances
    in table 4a and b, so in this way continuing to use monte carlo is just enlarging uncertainly.

    I think with many random parameter sets, the uncertainties determined by the monte carlo method should display variance
    that can be calculated by linear method.
    """

    print("\n========== result 3 ==========\n")

    res = ap.files.basic.read("results3.txt")
    le = [i[0] for i in res]
    lb = [i[1] for i in res]
    k = [i[2] for i in res]

    print(np.cov(le, k))
    print(np.cov(le, k)[1][0] / (np.std(le) * np.std(k)))

    print(f"LE = " + '{0} ± {1}'.format(*get_statistic(le)[:2]))
    print(f"LB = " + '{0} ± {1}'.format(*get_statistic(lb)[:2]))
    print(f"K = " + '{0} ± {1}'.format(*get_statistic(k)[:2]))
    print(f"age = " + '{0} ± {1}'.format(*calc_age(R, sR, le, lb, k)[:2]))


    print("\n========== result 2 ==========\n")

    res = ap.files.basic.read("results2.txt")
    le = [i[0] for i in res]
    lb = [i[1] for i in res]
    k = [i[2] for i in res]

    print(np.cov(le, k))
    print(np.cov(le, k)[1][0] / (np.std(le) * np.std(k)))

    print(f"LE = " + '{0} ± {1}'.format(*get_statistic(le)[:2]))
    print(f"LB = " + '{0} ± {1}'.format(*get_statistic(lb)[:2]))
    print(f"K = " + '{0} ± {1}'.format(*get_statistic(k)[:2]))
    print(f"age = " + '{0} ± {1}'.format(*calc_age(R, sR, le, lb, k)[:2]))


    print("\n========== result 1 ==========\n")

    res = ap.files.basic.read("results.txt")
    le = [i[0] for i in res]
    lb = [i[1] for i in res]
    k = [i[2] for i in res]

    print(np.cov(le, k))
    print(np.cov(le, k)[1][0] / (np.std(le) * np.std(k)))

    print(f"LE = " + '{0} ± {1}'.format(*get_statistic(le)[:2]))
    print(f"LB = " + '{0} ± {1}'.format(*get_statistic(lb)[:2]))
    print(f"K = " + '{0} ± {1}'.format(*get_statistic(k)[:2]))
    print(f"age = " + '{0} ± {1}'.format(*calc_age(R, sR, le, lb, k)[:2]))


def monte_carlo_to_calc_uncertainty():

    R, sR = 50, 0.25

    # print("\n========== result 1 ==========\n")
    #
    # res = ap.files.basic.read("results.txt")
    # le = [i[0] for i in res]
    # lb = [i[1] for i in res]
    # k = [i[2] for i in res]
    #
    # sle = np.std(le)
    # slb = np.std(lb)
    # sk = np.std(k)
    #
    # print(f"Std of LE  = {sle}")
    # print(f"Std of LB  = {slb}")
    # print(f"Std of K  = {sk}")
    #
    # cov_le_lb = np.cov(le, lb)[1][0]
    # cov_le_k = np.cov(le, k)[1][0]
    # cov_lb_k = np.cov(lb, k)[1][0]
    #
    # covariance_matrix = np.array(
    #     [[sle**2, cov_le_lb, cov_le_k],
    #      [cov_le_lb, slb**2, cov_lb_k],
    #      [cov_le_k, cov_lb_k, sk**2]]
    # )
    covariance_matrix = np.array(
        [[(0.0016E-10)**2, -3.4497E-26, 7.1889E-19],
         [-3.4497E-26, (0.0093E-10)**2, -3.4497E-26],
         [7.1889E-19, -3.4497E-26, (0.0045E-3)**2]]
    )

    print(covariance_matrix)

    ages, sages = [], []
    for i in range(4000):
        random_numbers = np.random.multivariate_normal(mean=[0.5755E-10, 4.9739E-10, 1.6418E-3], cov=covariance_matrix, size=4000).transpose()
        try:
            age, sage = calc_age(R, sR, random_numbers[0], random_numbers[1], random_numbers[2])[:2]
        except ValueError:
            continue
        else:
            ages.append(age)
            sages.append(sage)
        if i % 100 == 0:
            print(f"age = " + '{0} ± {1}'.format(str(age), str(sage)))

    print("\n========== final age ==========\n")

    print(f"age = " + '{0} ± {1}'.format(str(np.mean(ages)), str(np.std(ages))))
    print(f"sage = " + '{0} ± {1}'.format(str(np.mean(sages)), str(np.std(sages))))

    print(f"cov = \n{np.cov(ages, sages)}")

    # == == == == == final age == == == == ==
    #
    # age = 1050.7300017256905 ± 0.06448789398340359
    # sage = 4.016417903455615 ± 0.04467695074911889
    # cov =
    # [[4.15972840e-03 - 2.25391959e-05]
    #  [-2.25391959e-05  1.99652906e-03]]

    with open("ages.txt", 'w') as f:  # save serialized json data to a readable text
        f.writelines("\n".join([str(i) for i in ages]))

    with open("sages.txt", 'w') as f:  # save serialized json data to a readable text
        f.writelines("\n".join([str(i) for i in sages]))


def ellipse():
    age = [1050.7300017256905, 0.06448789398340359]
    sage = [4.016417903455615, 0.04467695074911889]
    cov = [[4.15972840E-03, -2.25391959E-05], [-2.25391959E-05, 1.99652906E-03]]
    ellipse_point = ap.calc.isochron.get_ellipse(*age, *sage, cov[0][1] * age[1] * sage[1], plt_sfactor=0, size=60)

    with open("ellipse.txt", 'w') as f:  # save serialized json data to a readable text
        f.writelines("\n".join([str(i) for i in ["{0}, {1}".format(str(j[0]), str(j[1])) for j in ellipse_point]]))

    print(ellipse_point)


if __name__ == "__main__":

    x0 = [
        LE, LB, K,
        6.789E-05,
        0.60556, 4.5494, 4.8205, 7.4636, 9.0619, 9.4983, 9.4918,
        9.5102, 9.4958, 9.5295, 9.5391, 9.6987, 12.6039, 18.09, 52.9011, 135.6389
    ]

    MF_regerssion([
        LE, LB, K,
        6.789E-05,
        0.60556, 4.5494, 4.8205, 7.4636, 9.0619, 9.4983, 9.4918,
        9.5102, 9.4958, 9.5295, 9.5391, 9.6987, 12.6039, 18.09, 52.9011, 135.6389
    ])

    print("\n ======== ti 最优值 （由参数计算而来） ======== \n")
    print([get_ti(Opt_LE, Opt_LB, Opt_K, ri) for ri in Opt_R])


    print("\n ======== MONTE CARLO 模拟 ======== \n")
    res = []
    num = 4000
    count = 0
    random_T = [np.random.normal(t, sT[i], num) for i, t in enumerate(T)]
    random_R = [np.random.normal(r, sR[i], num) for i, r in enumerate(R)]
    random_LE = np.random.normal(LE, sLE, num)
    random_LB = np.random.normal(LB, sLB, num)
    random_LT = np.random.normal(LT, sLT, num)
    random_K = np.random.normal(K, sK, num)

    while count < num:

        T = [random_T[i][count] for i, _ in enumerate(T)]
        R = [random_R[i][count] for i, _ in enumerate(R)]
        LE = random_LE[count]
        LB = random_LB[count]
        LT = random_LT[count]
        K = random_K[count]
        if count % 100 == 0:
            print(count)
        count += 1

        a = scipy.optimize.minimize(fun=func, x0=x0, method="Nelder-Mead", options={"maxiter": 10000}, tol=0.00000000001)
        res.append(a["x"].tolist())

    print("\n ======== 参数最优结果 ======== \n")
    print(res)

    ap.files.basic.write("results-0213.txt", res)

    # this_is_to_identify_that_monte_carlo_is_not_better_than_linear()
    # monte_carlo_to_calc_uncertainty()
    # ellipse()
    pass


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - age
# ==========================================
#
#
#
"""

# === Internal import ===
from . import arr

# === External import ===
import numpy as np


def calc_age_min(F, sF, **kwargs) -> tuple:
    """
    Calculate ArAr ages using Min et al. (2000) equation.
    Note that the Ar/K of primary standards is not working currently
    Parameters
    ----------
    F : array-like
    sF : array-like
    kwargs :
         keys : [
            'F', 'sF', 'L', 'sL', 'Le', 'sLe', 'Lb', 'sLb', 'A', 'sA', 'Ae', 'sAe', 'Ab', 'sAb',
            't', 'st', 'J', 'sJ', 'W', 'sW', 'No', 'sNo', 'Y', 'sY', 'f', 'sf', 'Min'
        ].
        values :  array-like objects.

    Returns
    -------
    [age, s1, s2, s3] in years

    """
    F, sF = arr.array_as_float([F, sF]).astype(float)
    J = arr.array_as_float(kwargs.pop('J'))
    sJ = arr.array_as_float(kwargs.pop('sJ') * J / 100)
    A = arr.array_as_float(kwargs.pop('A'))
    sA = arr.array_as_float(kwargs.pop('sA') * A / 100)
    Ae = arr.array_as_float(kwargs.pop('Ae'))
    sAe = arr.array_as_float(kwargs.pop('sAe') * Ae / 100)
    Ab = arr.array_as_float(kwargs.pop('Ab'))
    sAb = arr.array_as_float(kwargs.pop('sAb') * Ab / 100)
    W = arr.array_as_float(kwargs.pop('W'))
    sW = arr.array_as_float(kwargs.pop('sW') * W / 100)
    Y = arr.array_as_float(kwargs.pop('Y'))
    sY = arr.array_as_float(kwargs.pop('sY') * Y / 100)
    f = arr.array_as_float(kwargs.pop('f'))
    sf = arr.array_as_float(kwargs.pop('sf') * f / 100)
    No = arr.array_as_float(kwargs.pop('No'))
    sNo = arr.array_as_float(kwargs.pop('sNo') * No / 100)
    L = arr.array_as_float(kwargs.pop('L'))
    sL = arr.array_as_float(kwargs.pop('sL') * L / 100)
    Le = arr.array_as_float(kwargs.pop('Le'))
    sLe = arr.array_as_float(kwargs.pop('sLe') * Le / 100)
    Lb = arr.array_as_float(kwargs.pop('Lb'))
    sLb = arr.array_as_float(kwargs.pop('sLb') * Lb / 100)
    t = arr.array_as_float(kwargs.pop('t'))
    st = arr.array_as_float(kwargs.pop('st') * t / 100)
    # Ap = arr.array_as_float(kwargs.pop('Ap'))
    # sAp = arr.array_as_float(kwargs.pop('sAp') * Ap / 100)
    # Kp = arr.array_as_float(kwargs.pop('Kp'))
    # sKp = arr.array_as_float(kwargs.pop('sKp') * Kp / 100)

    # recalculation using Min et al.(2000) equation
    # lmd = A * W * Y / (f * No)
    V = f * No / ((Ab + Ae) * W * Y)
    sV = pow((V / f * sf) ** 2 + (V / No * sNo) ** 2 + (V / (Ab + Ae)) ** 2 * (sAb ** 2 + sAe ** 2) +
             (V / W * sW) ** 2 + (V / Y * sY) ** 2, 0.5)
    # standard age in year, change to Ma
    t = t * 1000000
    st = st * 1000000
    # back-calculating Ar40/Ar39 ration for the standard
    stdR = (np.exp(t * L) - 1) / J
    # errors of standard age and decay constants were not applied
    sStdR = pow((stdR / J) ** 2 * sJ ** 2, 0.5)
    # normalize the measured 40Ar/39Ar
    R = F / stdR
    sR_1 = np.sqrt((sF / stdR) ** 2 + (F * sStdR / stdR ** 2) ** 2)  # errors of measured 40Ar/39Ar and J value
    sR_2 = np.sqrt((sF / stdR) ** 2)  # error of measured 40Ar/39Ar only

    BB = 1
    KK = np.exp(t / V) - 1  # 40Arr / 40K Use standard age
    XX = BB * KK * R + 1
    k0 = V * np.log(XX)
    e1 = (np.log(XX) * V / f - V * BB * KK * R / (f * XX)) ** 2 * sf ** 2  # sF
    e2 = (np.log(XX) * V / No) ** 2 * sNo ** 2  # sNo
    e3 = (-1 * np.log(XX) * V / A + BB * KK * R / (A * XX)) ** 2 * sAb ** 2  # sAb
    e4 = (-1 * np.log(XX) * V / A - Ab * KK * R / (Ae ** 2 * XX)) ** 2 * sAe ** 2  # sAe
    e5 = (-1 * np.log(XX) * V / W - V * BB * KK * R / (W * XX)) ** 2 * sW ** 2  # sW
    e6 = (np.log(XX) * V / Y) ** 2 * sY ** 2  # sY
    e7 = (V * BB * KK / XX) ** 2 * sR_1 ** 2  # sR
    # e8 = (V * BB * KK * R / (Ap * XX)) ** 2 * sAp ** 2  # sAp
    # e9 = (V * BB * KK * R / (Kp * XX)) ** 2 * sKp ** 2  # sKp
    e8, e9 = 0, 0
    # useDecayConst = False
    # if useDecayConst:  # k0 = log(L / Le * KK * R + 1) / L
    #     e1 = (V * BB * KK * R / (f * XX)) ** 2 * sf ** 2
    #     e2 = 0
    #     e3 = (-1 * np.log(XX) * V / L + BB * KK * R / (L * XX)) ** 2 * sLb ** 2
    #     e4 = (-1 * np.log(XX) * V / L - Lb * KK * R / (Le ** 2 * XX)) ** 2 * sLe ** 2
    #     e5 = (V * BB * KK * R / (W * XX)) ** 2 * sW ** 2
    #     e6 = 0
    useStandardAge = True
    if useStandardAge:
        e1, e9 = 0, 0

    # change to Ma
    # analytical error, error of 40Ar/39Ar only
    s1 = np.sqrt((V * KK * BB / XX) ** 2 * sR_2 ** 2)
    # internal error, errors of 40Ar/39Ar and J value
    s2 = np.sqrt((V * KK * BB / XX) ** 2 * sR_1 ** 2)
    # total external error
    s3 = np.sqrt(e1 + e2 + e3 + e4 + e5 + e6 + e7 + e8 + e9)
    age = k0
    return age, s1, s2, s3


def calc_age_general(F, sF, J, sJ, L, sL, **kwargs):
    """
    All passed args should be array-like with the same length
    Parameters
    ----------
    F : array-like object
    sF : array-like object
    J : array-like object
    sJ : array-like object
    L : array-like object
    sL : array-like object
    kwargs

    Returns
    -------
    tuple of array objects, (age, s1, s2, s3), in years
    """
    F, sF, J, sJ, L, sL = np.array([F, sF, J, sJ, L, sL])
    sJ = sJ * J / 100
    sL = sL * L / 100
    age = np.log(1 + J * F) / L
    v1 = sF ** 2 * (J / (L * (1 + J * F))) ** 2
    v2 = sJ ** 2 * (F / (L * (1 + J * F))) ** 2
    v3 = sL ** 2 * (np.log(1 + J * F) / (L ** 2)) ** 2
    s1 = v1 ** .5  # analytical error
    s2 = (v1 + v2) ** .5  # internal error
    s3 = (v1 + v2 + v3) ** .5  # full external error
    return age, s1, s2, s3



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc- spectra
# ==========================================
#
# Spectra
#
"""
import numpy as np


def get_data(y: list, sy: list, x: list, f: int = 1, indices: list = None,
             cumulative: bool = False, successive: bool = True, sigma: int = 1):
    """
    Get spectra data based on passed x, y, and sy.

    Parameters
    ----------
    y :
    sy :
    x :
    f : error factor of sy, 1 or 2 for 1 or 2 sigma
    indices : array-like, default None for all sequences included
        An array of ints indicating which sequences to include.
    cumulative : bool, default False.
        This parameter should be True if x is already cumulative.
    successive : If setting indices successive
    sigma:

    Returns
    -------
    list of lists of float. [x, y1, y2]
    """
    if np.issubdtype(type(f), np.integer) and f > 1:
        sy = np.divide(sy, f)
    sy = np.array(sy) * sigma
    dp = np.shape([y, sy, x])[-1]
    if indices is None:
        indices = list(range(dp))
    if successive:
        indices = list(range(min(indices), max(indices) + 1))
    if not cumulative:
        x = np.divide(np.cumsum(x) * 100, sum(x)).tolist()
    k0, k1, k2 = [], [], []
    for i in range(dp):
        if i not in indices:
            continue
        else:
            k0.append([x[i - 1], 0][i == 0])
            k0.append(x[i])
            k1.append(y[i] + sy[i] if i % 2 == 0 else y[i] - sy[i])
            k1.append(y[i] + sy[i] if i % 2 == 0 else y[i] - sy[i])
            k2.append(y[i] - sy[i] if i % 2 == 0 else y[i] + sy[i])
            k2.append(y[i] - sy[i] if i % 2 == 0 else y[i] + sy[i])
    # Close the spectrum
    k0.insert(0, k0[0])
    k0.append(k0[-1])
    k1.insert(0, k2[0])
    k1.append(k2[-1])
    k2.insert(0, k1[0])
    k2.append(k1[-1])
    return [k0, k1, k2]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - plot
# ==========================================
#
#
#
"""
import traceback

import decimal
from math import exp, log, cos, acos, ceil, sqrt, atan, sin, gamma
# from typing import List, Any
# from scipy.optimize import fsolve
# from scipy.stats import distributions
import numpy as np

math_e = 2.718281828459045
math_pi = 3.141592653589793


def get_axis_scale(data: list, count=6, increment=None, extra_count=0, min_interval=1):
    """

    Parameters
    ----------
    data
    count
    increment
    extra_count
    min_interval

    Returns
    -------

    """
    if len(data) == 0:
        return 0, 1, 5, 0.2  # Default return
    _max = float(np.ma.masked_invalid(data).max())
    _min = float(np.ma.masked_invalid(data).min())
    interval = (_max - _min) / count
    if interval == 0:
        interval = 10
    mag = 10 ** int(log(interval, 10)) if interval >= 1 else 10 ** (int(log(interval, 10)) - 1)
    if not increment:
        increment = decimal.Decimal(
            str(decimal.Decimal(int(interval / mag // min_interval) + min_interval) * decimal.Decimal(str(mag))))
    else:
        increment = decimal.Decimal(increment)
    start = decimal.Decimal(0)
    if _min < 0:
        start = -increment
        while start > _min:
            start -= increment
    else:
        while start + increment < _min:
            start += increment
    count = 0
    while count * increment + start < _max:
        count += 1
    end = decimal.Decimal(str(count + extra_count)) * decimal.Decimal(str(increment)) + start
    start -= decimal.Decimal(extra_count) * decimal.Decimal(str(increment))
    start = 0 if start < 0 <= _min else start
    count = (end - start) / increment
    return float(start), float(end), int(count), float(increment)


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
"""
# ==========================================
# Copyright 2023 Yang
# ararpy - calc - regression
# ==========================================
#
# Regression functions
#
"""
from . import arr

# === External import ===
import traceback
import numpy as np
import pandas as pd
from scipy.stats import distributions
from scipy.optimize import fsolve
import warnings
from scipy.optimize import minimize
warnings.simplefilter(action="ignore", category=RuntimeWarning)


""" regression functions for isochrons """


def york2(x: list, sx: list, y: list, sy: list, ri: list, f: int = 1,
          convergence: float = 0.001, iteration: int = 100):
    """
    Parameters
    ----------
    x : isochron x-axis
    sx : standard error of x
    y : isochron y-axis, y = b + m * x
    sy : standard error of y
    ri : error correlation coefficient of errors of x and y
    f : factor of errors, default 1
    convergence: float. Convergence tolerate, default 0.001
    iteration: int. Number of iteration, default 100

    Returns
    -------
    Intercept | Error | slope | Error | MSWD | Convergence | Number of Iterations | error magnification | other
    b, sb, a, sa, mswd, dF, Di, k, r2, chi_square, p_value, avg_err_s, cov_b_m
    b, seb, m, sem, mswd, abs(m - last_m), Di, k, r2, chi_square, p_value, avg_err_s, cov_b_m
    """
    data = np.array([x, sx, y, sy, ri])
    data = data[:, np.where(
        np.logical_or(data == np.inf, pd.isna(data)), False, True).all(axis=0)].astype(np.float64)
    x, sx, y, sy, ri = data
    n = data.shape[-1]
    X, sX, Y, sY, R = data
    # change to 1 sigma
    if np.issubdtype(type(f), np.integer) and f > 1:
        sX, sY = np.divide([sX, sY], f)
    # weights of x and y
    wX = 1 / sX ** 2
    wY = 1 / sY ** 2
    # weight of S
    Z = lambda m, b: wX * wY / (m ** 2 * wY + wX - 2 * m * R * (wX * wY) ** .5)
    # weighted mean of X and Y
    mX = lambda m, b: sum(Z(m, b) * X) / sum(Z(m, b))
    mY = lambda m, b: sum(Z(m, b) * Y) / sum(Z(m, b))
    # Equation to minimize
    S = lambda m, b: sum(Z(m, b) * (Y - m * X - b) ** 2)
    # Slope by OLS is used as the initial values in weights calculation
    temp_lst = linest(Y, X)
    if not temp_lst:
        return False
    b, seb, m, sem = temp_lst[5][0], temp_lst[6][0], temp_lst[5][1], temp_lst[6][1]
    b = mY(m, b) - m * mX(m, b)
    last_m = 1e10
    Di = 0  # Iteration number
    mswd, k = 1, 1  # Initial return values
    while abs(m - last_m) >= abs(m * convergence / 100):
        last_m = m
        U = X - mX(m, b)
        V = Y - mY(m, b)
        # Expression from York 2004, which differs to York 1969
        Up = Z(m, b) ** 2 * V * (U / wY + m * V / wX - R * (V + m * U) / (wX * wY) ** .5)
        Lo = Z(m, b) ** 2 * U * (U / wY + m * V / wX - R * (V + m * U) / (wX * wY) ** .5)
        m = sum(Up) / sum(Lo)  # New slope
        b = mY(m, b) - m * mX(m, b)  # From York 2004, calculate b again after final value of m has been obtained
        sumUUZ = sum(U * U * Z(m, b))
        sumXXZ = sum(X * X * Z(m, b))
        sem = 1 / sumUUZ ** .5
        seb = (sumXXZ / sum(Z(m, b))) ** .5 * sem
        mswd = S(m, b) / (n - 2)
        # print(f"York 2004 regression, m = {m}, b = {b}, S = {S(m, b)}, Di = {Di}")
        if mswd > 1:
            k = mswd ** .5  # k为误差放大系数
        else:
            k = 1

        sem = sem * k
        seb = seb * k

        Di = Di + 1
        if Di >= iteration:
            break

    # Calculate Y values base on the regression results
    estimate_y = b + m * X
    resid = (estimate_y - Y) ** 2
    reg = (estimate_y - np.mean(estimate_y)) ** 2
    ssresid = sum(resid)  # residual sum of squares / sum squared residual
    ssreg = sum(reg)  # regression sum of square
    sstotal = ssreg + ssresid  # total sum of squares
    r2 = ssreg / sstotal if sstotal != 0 else np.inf  # r2 = ssreg / sstotal
    chi_square = mswd * (n - 2)
    p_value = distributions.chi2.sf(chi_square, n - 2)
    cov_b_m = - np.mean(X) * (ssresid / (n - 2) / np.sum((X - np.mean(X)) ** 2))  # covariance of intercept b and slope m
    # average error of S
    err_s = lambda m, b: list(map(lambda Zi, Yi, Xi: (1 / Zi) ** (1./2.) / abs(Yi - m * Xi - b), Z(m, b), y, x))
    avg_err_s = sum(err_s(m, b)) / len(x) * 100

    # print('----------------------------------------------------------------')
    # print('截距>>>' + str(b) + '  ' + '误差>>>' + str(seb))
    # print('斜率>>>' + str(m) + '  ' + '误差>>>' + str(sem))
    # print('Absolute Convergence' + '>>>' + str(abs(m - last_m)))
    # print('Number of Iterations' + '>>>' + str(Di))
    # print('MSWD' + '>>>' + str(mswd))
    # print('Error Magnification>>>' + str(k))
    # print('----------------------------------------------------------------')

    # keys = [
    #     k, sk, m, sm, mswd, conv, iter, mag, r2, chisq, p, avg_err
    # ]

    return b, seb, m, sem, mswd, abs(m - last_m), Di, k, r2, chi_square, p_value, avg_err_s, cov_b_m


def york2_df(data: pd.DataFrame, f: int = 1, convergence: float = 0.001,
             iteration: int = 100) -> pd.DataFrame:
    """
    DataFrame format input and output
    Parameters
    ----------
    data
    f
    convergence
    iteration

    Returns
    -------
    data frame with keys [
            'k', 'sk', 'm1', 'sm1',
            'iter', 'conv', 'rs', 'MSWD', 'R2', 'Chisq', 'Pvalue',  # 'rs' means relative error of the total sum
            'mag', 'abs_conv',  # 'mag' is error magnification factor, 'abs_conv' is absolute convergence at the last time
        ]
    Intercept | Error | slope | Error | MSWD | Convergence | Number of Iterations | error magnification | other
         b, sb, a, sa, mswd, dF, Di, k, r2, chi_square, p_value
    """
    # remove nan, inf from data
    data: pd.DataFrame = data.replace([-np.inf], np.nan).dropna(axis=0)
    data: list = data.transpose().values.tolist()[:5]

    columns = [
        'k', 'sk', 'm1', 'sm1',
        'MSWD', 'abs_conv', 'iter', 'mag', 'R2', 'Chisq', 'Pvalue',  # 'rs' means relative error of the total sum
        'rs'
    ]
    values = york2(*data, f=f, convergence=convergence, iteration=iteration)
    res = pd.DataFrame([list(values)], columns=columns)

    return res


def wtd_3D_regression(x: list, sx: list, y: list, sy: list, z: list, sz: list, r1: list,
                      r2: list, r3: list, f: int = 1, convergence: float = 0.001,
                      iteration: int = 100):
    """
    Error weighted regression for 3D plots
    Parameters
    ----------
    x
    sx
    y
    sy
    z
    sz
    r1 : error correlation between x and y
    r2 : error correlation between x and z
    r3 : error correlation between y and z
    f : factor of errors, default 1.
    convergence : convergence tolerate in percentage, default 0.001 indicating 0.001%
    iteration : number of iteration, default 100

    Returns
    -------
    c (interceept), sc, a, sa, b, sb, S, mswd, r2, abs(a - last_a), Di, k  # length == 12
    """
    n = np.shape([x, sx, y, sy, z, sz, r1, r2, r3])[-1]
    x, sx, y, sy, z, sz, r1, r2, r3 = np.array([x, sx, y, sy, z, sz, r1, r2, r3])
    # change to 1 sigma
    if np.issubdtype(type(f), np.integer) and f > 1:
        sx, sy, sz = np.divide([sx, sy, sz], f)
    if n <= 3:
        return False

    Di = 0

    # Weights of S
    W = lambda a, b: 1 / (
                a ** 2 * sx ** 2 + b ** 2 * sy ** 2 + sz ** 2 + 2 * a * b * r1 * sx * sy -
                2 * a * r2 * sx * sz - 2 * b * r3 * sy * sz)
    # Weighted mean values of X, Y, and Z, respectively
    mX = lambda a, b: sum(W(a, b) * x) / sum(W(a, b))
    mY = lambda a, b: sum(W(a, b) * y) / sum(W(a, b))
    mZ = lambda a, b: sum(W(a, b) * z) / sum(W(a, b))
    # Minimizing this equation
    S = lambda a, b, c: sum(W(a, b) * (a * x + b * y + c - z) ** 2)
    # Calculate new c based on iterated a and b
    new_c = lambda a, b: mZ(a, b) - a * mX(a, b) - b * mY(a, b)
    # Initial values of a, b, and c from OLS
    linest_res = linest(z, x, y)
    c, sc, k2, k3, k4, [_, a, b], [_, sa, sb] = linest_res[0:7]
    c = new_c(a, b)
    k = 1  # Error magnification factor
    last_a = 1e10
    mswd, f = 1000, 0
    # print(f"初始值：a = {a}, b = {b}, c = {c}")
    # ar38ar36 = 0.1885
    # ar40ar36 = (a + b * ar38ar36) * -1 / c
    # print(f"Ar38/Ar36 = {ar38ar36}, Ar40/Ar36 = {ar40ar36}, S = {S(a, b, c)}")
    while abs(a - last_a) >= abs(a * convergence / 100):
        last_a = a
        U = x - mX(a, b)
        V = y - mY(a, b)
        G = z - mZ(a, b)
        # P and Q are Xi - mX and Yi - mY, respectively. These values are obtained by weighted Orthogonal regression
        P = W(a, b) * ((a * sx ** 2 + b * r1 * sx * sy - r2 * sx * sz) * (G - b * V) + (
                    a * b * r1 * sx * sy + b ** 2 * sy ** 2 - a * r2 * sx * sz - 2 * b * r3 * sy * sz + sz ** 2) * U)
        Q = W(a, b) * ((b * sy ** 2 + a * r1 * sx * sy - r3 * sy * sz) * (G - a * U) + (
                a * b * r1 * sx * sy + a ** 2 * sx ** 2 - b * r3 * sy * sz - 2 * a * r2 * sx * sz + sz ** 2) * V)
        a_Up = sum(W(a, b) * P * G).sum() * sum(W(a, b) * Q * V).sum() - \
               sum(W(a, b) * P * V).sum() * sum(W(a, b) * Q * G)
        a_Lo = sum(W(a, b) * P * U).sum() * sum(W(a, b) * Q * V).sum() - \
               sum(W(a, b) * P * V).sum() * sum(W(a, b) * Q * U)
        new_a = a_Up / a_Lo
        b_Up = sum(W(a, b) * Q * G) * sum(W(a, b) * P * U) - sum(W(a, b) * P * G) * sum(W(a, b) * Q * U)
        b_Lo = sum(W(a, b) * P * U) * sum(W(a, b) * Q * V) - sum(W(a, b) * P * V) * sum(W(a, b) * Q * U)
        new_b = b_Up / b_Lo

        # Standard errors
        mU = sum(W(a, b) * U) / sum(W(a, b))
        mV = sum(W(a, b) * V) / sum(W(a, b))
        mP = sum(W(a, b) * P) / sum(W(a, b))
        mQ = sum(W(a, b) * Q) / sum(W(a, b))

        D_PU = W(a, b) * (a * b * r1 * sx * sy + b ** 2 * sy ** 2 - a * r2 * sx * sz - 2 * b * r3 * sy * sz + sz ** 2)
        D_QU = -1 * a * W(a, b) * (b * sy ** 2 + a * r1 * sx * sy - r3 * sy * sz)
        D_PV = -1 * b * W(a, b) * (a * sx ** 2 + b * r1 * sx * sy - r2 * sx * sz)
        D_QV = W(a, b) * (a * b * r1 * sx * sy + a ** 2 * sx ** 2 - b * r3 * sy * sz - 2 * a * r2 * sx * sz + sz ** 2)
        D_PG = W(a, b) * (a * sx ** 2 + b * r1 * sx * sy - r2 * sx * sz)
        D_QG = W(a, b) * (b * sy ** 2 + a * r1 * sx * sy - r3 * sy * sz)
        D_UX = D_VY = D_GZ = 1 - W(a, b) / sum(W(a, b))
        D_Wa = -1 * W(a, b) ** 2 * (2 * a * sx ** 2 + 2 * b * r1 * sx * sy - 2 * r2 * sx * sz)
        D_Wb = -1 * W(a, b) ** 2 * (2 * b * sy ** 2 + 2 * a * r1 * sx * sy - 2 * r3 * sy * sz)

        D_aX = W(a, b) * D_UX * (a * (sum(W(a, b) * P * U) * V * D_QU + sum(W(a, b) * Q * V) * (
                U * D_PU + P) - sum(W(a, b) * Q * U) * V * D_PU - sum(W(a, b) * P * V) * (U * D_QU + Q)) -
                                 (sum(W(a, b) * P * G) * V * D_QU + sum(W(a, b) * Q * V) * G * D_PU) +
                                 (sum(W(a, b) * Q * G) * V * D_PU + sum(W(a, b) * P * V) * G * D_QU))

        D_aY = W(a, b) * D_VY * (a * (sum(W(a, b) * P * U) * (Q + V * D_QV) + sum(W(a, b) * Q * V) * (
                U * D_PV) - sum(W(a, b) * Q * U) * (P + V * D_PV) - sum(W(a, b) * P * V) * (U * D_QV)) -
                                 (sum(W(a, b) * P * G) * (Q + V * D_QV) + sum(W(a, b) * Q * V) * G * D_PV) +
                                 (sum(W(a, b) * Q * G) * (P + V * D_PV) + sum(W(a, b) * P * V) * G * D_QV))


        D_aZ = W(a, b) * D_GZ * (a * (sum(W(a, b) * P * U) * (V * D_QG) + sum(W(a, b) * Q * V) * (
                U * D_PG) - sum(W(a, b) * Q * U) * (V * D_PG) - sum(W(a, b) * P * V) * (U * D_QG)) -
                                 (sum(W(a, b) * P * G) * (V * D_QG) + sum(W(a, b) * Q * V) * (P + G * D_PG)) +
                                 (sum(W(a, b) * Q * G) * (V * D_PG) + sum(W(a, b) * P * V) * (Q + G * D_QG)))

        D_WPU_a = D_Wa * P * U
        D_WQV_a = D_Wa * Q * V
        D_WQU_a = D_Wa * Q * U
        D_WPV_a = D_Wa * P * V
        D_WPG_a = D_Wa * P * G
        D_WQG_a = D_Wa * Q * G

        D_aa = a_Lo + \
               a * (sum(D_WPU_a) * sum(W(a, b) * Q * V) + sum(D_WQV_a) * sum(W(a, b) * P * U) -
                    sum(D_WQU_a) * sum(W(a, b) * P * V) - sum(D_WPV_a) * sum(W(a, b) * Q * U)
        ) - (sum(D_WPG_a) * sum(W(a, b) * Q * V) + sum(D_WQV_a) * sum(W(a, b) * P * G) -
             sum(D_WQG_a) * sum(W(a, b) * P * V) - sum(D_WPV_a) * sum(W(a, b) * Q * G))

        D_bX = W(a, b) * D_UX * (b * (sum(W(a, b) * P * U) * (V * D_QU) + sum(W(a, b) * Q * V) * (P + U * D_PU) -
                                      sum(W(a, b) * Q * U) * (V * D_PU) - sum(W(a, b) * P * V) * (Q + U * D_QU)) -
                                 (sum(W(a, b) * Q * G) * (P + U * D_PU) + sum(W(a, b) * P * U) * G * D_QU) +
                                 (sum(W(a, b) * P * G) * (Q + U * D_QU) + sum(W(a, b) * Q * U) * G * D_PU))

        D_bY = W(a, b) * D_VY * (b * (sum(W(a, b) * P * U) * (Q + V * D_QV) + sum(W(a, b) * Q * V) * (U * D_PV) -
                                      sum(W(a, b) * Q * U) * (P + V * D_PV) - sum(W(a, b) * P * V) * (U * D_QV)) -
                                 (sum(W(a, b) * Q * G) * (U * D_PV) + sum(W(a, b) * P * U) * (G * D_QV)) +
                                 (sum(W(a, b) * P * G) * (U * D_QV) + sum(W(a, b) * Q * U) * (G * D_PV)))

        D_bZ = W(a, b) * D_GZ * (b * (sum(W(a, b) * P * U) * (V * D_QG) + sum(W(a, b) * Q * V) * (U * D_PG) -
                                      sum(W(a, b) * Q * U) * (V * D_PG) - sum(W(a, b) * P * V) * (U * D_QG)) -
                                 (sum(W(a, b) * Q * G) * (U * D_PG) + sum(W(a, b) * P * U) * (Q + G * D_QG)) +
                                 (sum(W(a, b) * P * G) * (U * D_QG) + sum(W(a, b) * Q * U) * (P + G * D_PG)))

        D_WPU_b = D_Wb * P * U
        D_WQV_b = D_Wb * Q * V
        D_WQU_b = D_Wb * Q * U
        D_WPV_b = D_Wb * P * V
        D_WPG_b = D_Wb * P * G
        D_WQG_b = D_Wb * Q * G

        D_bb = b_Lo + b * (
                sum(D_WPU_b) * sum(W(a, b) * Q * V) + sum(D_WQV_b) * sum(W(a, b) * P * U) -
                sum(D_WQU_b) * sum(W(a, b) * P * V) - sum(D_WPV_b) * sum(W(a, b) * Q * U)
        ) - (
                sum(D_WQG_b) * sum(W(a, b) * P * U) + sum(D_WPU_b) * sum(W(a, b) * Q * G) -
                sum(D_WPG_b) * sum(W(a, b) * Q * U) - sum(D_WQU_b) * sum(W(a, b) * P * G)
               )

        Va = sum(D_aX ** 2 * sx ** 2 + D_aY ** 2 * sy ** 2 + D_aZ ** 2 * sz ** 2 +
                 2 * r1 * sx * sy * D_aX * D_aY + 2 * r2 * sx * sz * D_aX * D_aZ + 2 * r3 * sy * sz * D_aY * D_aZ)
        Vb = sum(D_bX ** 2 * sx ** 2 + D_bY ** 2 * sy ** 2 + D_bZ ** 2 * sz ** 2 +
                 2 * r1 * sx * sy * D_bX * D_bY + 2 * r2 * sx * sz * D_bX * D_bZ + 2 * r3 * sy * sz * D_bY * D_bZ)

        D_cX = - 1 * a * W(a, b) / sum(W(a, b)) + (-1 * D_aX) * (2 * mP - 2 * mU + mX(a, b)) + (-1 * D_bX) * (
                                2 * mQ - 2 * mV + mY(a, b))
        D_cY = - 1 * b * W(a, b) / sum(W(a, b)) + (-1 * D_aY) * (2 * mP - 2 * mU + mX(a, b)) + (-1 * D_bY) * (
                                2 * mQ - 2 * mV + mY(a, b))
        D_cZ = W(a, b) / sum(W(a, b)) + (-1 * D_aZ) * (2 * mP - 2 * mU) + (-1 * D_bZ) * (2 * mQ - 2 * mV)
        Vc = sum(D_cX ** 2 * sx ** 2 + D_cY ** 2 * sy ** 2 + D_cZ ** 2 * sz ** 2 +
                 2 * r1 * sx * sy * D_cX * D_cY + 2 * r2 * sx * sz * D_cX * D_cZ + 2 * r3 * sy * sz * D_cY * D_cZ)

        sa = (Va / D_aa) ** .5
        sb = (Vb / D_bb) ** .5
        sc = Vc ** .5

        mswd = S(a, b, c) / (n - 3)
        if mswd > 1:
            k = mswd ** .5  # k为误差放大系数
        else:
            k = 1

        sa, sb, sc = sa * k, sb * k, sc * k

        a = new_a
        b = new_b
        c = new_c(new_a, new_b)

        # ar40ar36 = (a + b * ar38ar36) * -1 / c
        # f = 1 / c
        # print(f"new_a = {a}, new_b = {b}, new_c = {c}, S = {S(a, b, c)}, MSWD = {mswd}, Ar38/Ar36 = {ar38ar36}, Ar40/Ar36 = {ar40ar36}")
        #
        # print(f"Iteration info: a = {a:.4f} ± {sa:.4f} | {sa/a * 100:.2f}%, b = {b:.4f} ± {sb:.4f} | {sb/b * 100:.2f}%, c = {c:.4f} ± {sc:.4f} | {sc/c * 100:.2f}% "
        #       f"S = {S(a, b, c)}， Di = {Di}, MSWD = {mswd}")

        Di = Di + 1
        if Di >= iteration:
            break

    estimate_z = c + a * x + b * y
    resid = (estimate_z - z) ** 2
    reg = (estimate_z - np.mean(estimate_z)) ** 2
    ssresid = sum(resid)  # residual sum of squares / sum squared residual
    ssreg = sum(reg)  # regression sum of square
    sstotal = ssreg + ssresid  # total sum of squares
    R = ssreg / sstotal if sstotal != 0 else np.inf  # r2 = ssreg / sstotal
    chi_square = mswd * (n - 3)
    p_value = distributions.chi2.sf(chi_square, n - 3)

    # relative error of S
    err_s = lambda a, b, c: (1 / W(a, b)) ** .5 / abs(a * x + b * y + c - z)
    avg_err_s = np.mean(err_s(a, b, c)) * 100
    # print(f"Average relative error of S = {avg_err_s}%")

    # print(f"a = {a}, b = {b}, c = {c}, S = {S(a, b, c)}， Di = {Di}, MSWD = {mswd}, r2 = {R}")

    return c, sc, a, sa, b, sb, S(a, b, c), mswd, R, abs(a - last_a), \
           Di, k, chi_square, p_value, avg_err_s


def wtd_3D_regression_df(data: pd.DataFrame, f: int = 1, convergence: float = 0.001,
                         iteration: int = 100) -> pd.DataFrame:
    """
    :param data: isochron data
    :param f: factor of error, should be 1 for 1 sigma, or 2 for 2 sigma, default = 1
    :param convergence: convergence toleration in percentage, default = 0.001, means 0.001%
    :param iteration number of iteration, default = 100
    :return: data frame with keys [
        'k', 'sk', 'm1', 'sm1',
        'iter', 'conv', 'rs', 'MSWD', 'R2', 'Chisq', 'Pvalue',  # 'rs' means relative error of the total sum
        'mag', 'abs_conv',  # 'mag' is error magnification factor, 'abs_conv' is absolute convergence at the last time
    ]
    """
    # remove nan, inf from data
    data: pd.DataFrame = data.replace([-np.inf], np.nan).dropna(axis=0)
    data = data.transpose().values.tolist()
    res_list = wtd_3D_regression(*data[:9], f=f, convergence=convergence, iteration=iteration)

    columns = [
        'k', 'sk', 'm1', 'sm1', 'm2', 'sm2',
        'S', 'MSWD', 'R2', 'abs_conv', 'iter', 'mag',
        'Chisq', 'Pvalue', 'rs',  # 'rs' means relative error of the total sum

    ]
    res = pd.DataFrame([list(res_list)], columns=columns)
    return res


def max_likelihood(x: list, sx: list, y: list, sy: list, ri: list):
    """
    Parameters
    ----------
    x
    sx
    y
    sy
    ri

    Returns
    -------

    """

    # 定义线性模型
    def linear_model(params, _x):
        return params[0] * _x + params[1]

    # 定义负对数似然函数
    def negative_log_likelihood(params, _x, _y):
        predicted_y = linear_model(params, _x)
        # 使用高斯误差进行似然计算
        errors = predicted_y - _y
        likelihood = np.sum(-0.5 * np.log(2 * np.pi) - 0.5 * errors ** 2)
        return -likelihood

    initial = linest(y, x)[5]
    res = minimize(negative_log_likelihood, initial, args=(np.array(x), np.array(y)))
    slope, intercept = res.x
    return intercept, 0, slope, 0, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan


def olst(x: list, sx: list, y: list, sy: list, ri: list):
    """
    Parameters
    ----------
    x
    sx
    y
    sy
    ri

    Returns
    -------

    """

    res = linest(y, x)
    # b, sb, m, sm, mswd, m - last m, Di, k, r2,
    return *res[:2], res[5][1], res[6][1], np.nan, np.nan, np.nan, np.nan, res[3], np.nan, np.nan, np.nan


""" regression functions for raw data """


def linest(a0: list, a1: list, *args):
    """
    Parameters
    ----------
    a0 : known_y's, y = b + m * x
    a1 : known_x's
    args : more known_x's

    Returns
    -------
    intercept | standard error | relative error | R2 | MSWD | other params: list |
             error of other params: list | equation | m_ssresid (y估计值的标准误差)

    """
    # beta = (xTx)^-1 * xTy >>> xtx * beta = xty
    # crate matrix of x and y, calculate the transpose of x
    if not args:
        x = np.concatenate(([[1]*len(a1)], [a1]), axis=0).transpose()
    else:
        x = np.concatenate(([[1]*len(a1)], [a1], args), axis=0).transpose()
    n = x.shape[-1]  # number of unknown x, constant is seen as x^0
    m = x.shape[0]  # number of data
    y = np.array([a0]).transpose()
    try:
        inv_xtx = np.linalg.inv(np.matmul(x.transpose(), x))
    except np.linalg.LinAlgError:
        raise np.linalg.LinAlgError(f"The determinant of the given matrix must not be zero ")
    beta = np.matmul(inv_xtx, np.matmul(x.transpose(), y))

    # calculate Y values base on the fitted formula
    estimate_y = np.matmul(x, beta)
    resid = (estimate_y - y) ** 2
    reg = (estimate_y - np.mean(estimate_y)) ** 2
    ssresid = resid.sum()
    ssreg = reg.sum()
    sstotal = ssreg + ssresid
    df = m - n
    m_ssresid = ssresid / df
    se_beta = (m_ssresid * np.diagonal(inv_xtx)) ** .5
    beta = beta.transpose()[0]
    rse_beta = se_beta / beta
    r2 = ssreg / sstotal if sstotal != 0 else np.inf

    def get_adjusted_y(*args):
        args = [[1] * len(args[0]), *args]
        return [sum([beta[i] * args[i][j] for i in range(len(beta))]) for j in range(len(args[0]))]

    return beta[0], se_beta[0], rse_beta[0] * 100, r2, m_ssresid, beta, se_beta, get_adjusted_y, m_ssresid


def average(a0: list, a1=None):
    """
    :param a0: known_y's
    :param a1:
    :return: intercept | standard error | relative error | r2 | MSWD | other params | errors of other params |
             euqation | m_ssresid
    """
    if a1 is None:
        a1 = []
    k0 = sum(a0) / len(a0)

    # calculate Y values base on the fitted formula
    estimate_y = [k0 for x in a0]
    resid = [(x - k0) ** 2 for x in a0]
    reg = [(i - sum(estimate_y) / len(estimate_y)) ** 2 for i in estimate_y]
    ssresid = sum(resid)  # residual sum of squares / sum squared residual
    ssreg = sum(reg)  # regression sum of square
    sstotal = ssreg + ssresid  # total sum of squares
    df = len(a0) - 1  # df = degree of freedom
    m_ssresid = ssresid / df
    r2 = ssreg / sstotal if sstotal != 0 else 1  # r2 = ssreg / sstotal

    k1 = pow(sum([(i - k0) ** 2 for i in a0]) / df, 0.5)  # standard deviation
    k2 = k1 / k0 * 100 if k0 != 0 else 0  # relative standard error
    k3 = r2  # determination coefficient
    k4 = 'MSWD'
    k5 = [k0]
    k6 = [k1]
    k8 = m_ssresid

    def get_adjusted_y(x: list):
        return [k0] * len(x)

    k7 = get_adjusted_y

    return k0, k1, k2, k3, k4, k5, k6, k7, k8


#
# def wtd_linest(a0: list, a1: list):
#     """
#     y = m * x + b,
#     :param a0: known_y's
#     :param a1: known_x's
#     :return: intercept | standard error | relative error | R2 | [m] | [sem]
#     """
#     linest_res = linest(a0, a1)
#     b0, seb0, rseb0, r2, mswd, [m0], [rem0] = linest_res[0:7]
#     y0 = list(map(lambda i: m0 * i + b0, a1))
#     resid = list(map(lambda i, j: i - j, y0, a0))
#     weight = list(map(lambda i: 1 / i ** 2, resid))  # Use weighting by inverse of the squares of residual
#
#     sum_wi = sum(weight)
#     sum_wiyi = sum(list(map(lambda i, j: i * j, weight, a0)))
#     sum_wixi = sum(list(map(lambda i, j: i * j, weight, a1)))
#     sum_wiyixi = sum(list(map(lambda i, j, g: i * j * g, weight, a0, a1)))
#     sum_wixixi = sum(list(map(lambda i, j, g: i * j * g, weight, a1, a1)))
#
#     m = (sum_wiyixi - sum_wixi * sum_wiyi / sum_wi) / (sum_wixixi - sum_wixi * sum_wixi / sum_wi)
#     b = (sum_wiyi - m * sum_wixi) / sum_wi
#     a0 = list(map(lambda i, j: i * j, weight, a0))
#     a1 = list(map(lambda i, j: i * j, weight, a1))
#     linest_res = intercept_linest(a0, a1, weight=weight)
#     b, seb, rseb, r2, mswd, [m], [sem] = linest_res[0:7]
#     return b, seb, rseb, r2, [m], [sem]
#
#
# def intercept_linest(a0: list, a1: list, *args, weight: list = None, interceptIsZero: bool = False):
#     """
#     :param a0: known_y's, y = b + m * x
#     :param a1: known_x's
#     :param args: more known_x's
#     :param weight: necessary when weighted least squares fitting
#     :param interceptIsZero: set b as zero, y = m * x
#     :return: intercept | standard error | relative error | R2 | MSWD | other params: list |
#              error of other params: list | equation | m_ssresid (y估计值的标准误差)
#     """
#     if interceptIsZero:
#         if len(a0) != len(a1) or len(args) > 0:
#             return False
#         try:
#             df = len(a0) - 1
#             m = sum(list(map(lambda x, y: x * y, a1, a0))) / sum(list(map(lambda x: x ** 2, a1)))
#             SSresid = sum(list(map(lambda x, y: y - x * m, a1, a0)))
#             sey = pow(SSresid / df, 0.5)
#             SSreg = sum(list(map(lambda x: (x * m) ** 2, a1)))
#             SStotal = SSreg + SSresid
#             R2 = SStotal / SSreg
#             sem = pow(SSresid / df * 1 / sum(list(map(lambda x: x ** 2, a1))), 0.5)
#             return m, sem, R2
#         except Exception:
#             return False
#     # beta = (xTx)^-1 * xTy >>> xtx * beta = xty
#     # crate matrix of x and y, calculate the transpose of x
#     m = len(a1)  # number of data
#     n = len(args) + 2  # number of unknown x, constant is seen as x^0
#     if m - n < 1 or len(a0) != len(a1):
#         return False
#     if weight is not None:
#         xlst = [weight, a1, *args]
#     else:
#         xlst = [[1] * m, a1, *args]
#     ylst = a0
#     xtx = list()
#     xty = list()
#     for i in range(n):
#         xtx.append([])
#         xty.append([])
#         xty[i] = sum([xlst[i][k] * ylst[k] for k in range(m)])
#         for j in range(n):
#             xtx[i].append([])
#             xtx[i][j] = sum([xlst[i][k] * xlst[j][k] for k in range(m)])
#     # solve the system of linear equations using LU factorization algorithm
#     # LU * beta = xty, U * beta = b, L * b = xty
#     l: List[List[Any]] = list()
#     u: List[List[Any]] = list()
#     b: List[Any] = list()
#     beta: List[Any] = list()
#     for i in range(n):
#         l.append([])
#         u.append([])
#         b.append([])
#         beta.append([])
#         for j in range(n):
#             l[i].append([])
#             u[i].append([])
#             if j > i:
#                 l[i][j] = 0
#             elif i > j:
#                 u[i][j] = 0
#             else:
#                 l[i][j] = 1
#     for i in range(n):
#         if i >= 1:
#             l[i][0] = xtx[i][0] / u[0][0]
#         for j in range(n):
#             if i == 0:
#                 u[i][j] = xtx[i][j]
#             elif i == 1 and j >= 1:
#                 u[i][j] = xtx[i][j] - l[i][0] * u[0][j]
#             elif i < n - 1:
#                 if j in range(1, i):
#                     l[i][j] = (xtx[i][j] - sum([l[i][r] * u[r][j] for r in range(j)])) / u[j][j]
#                 if j in range(i, n):
#                     u[i][j] = xtx[i][j] - sum([l[i][r] * u[r][j] for r in range(i)])
#             elif i == n - 1:
#                 if j in range(1, i):
#                     l[n - 1][j] = (xtx[n - 1][j] - sum([l[n - 1][r] * u[r][j] for r in range(j)])) / u[j][j]
#                 if j == n - 1:
#                     u[i][j] = xtx[i][j] - sum([l[i][r] * u[r][j] for r in range(i)])
#     # calculate matrix b, L * b = y
#     b[0] = xty[0]
#     for i in range(1, n):
#         b[i] = xty[i] - sum([l[i][j] * b[j] for j in range(i)])
#     # calculate matrix beta, b = U * beta
#     beta[n - 1] = b[n - 1] / u[n - 1][n - 1]
#     for i in [n - k for k in range(2, n + 1)]:
#         beta[i] = (b[i] - sum([u[i][j] * beta[j] for j in range(i + 1, n)])) / u[i][i]
#
#     # calculate the inverse of matrix xTx
#     inv_l: List[List[Any]] = list()
#     inv_u: List[List[Any]] = list()
#     for i in range(n):
#         inv_l.append([])
#         inv_u.append([])
#         for j in range(n):
#             inv_l[i].append([])
#             inv_u[i].append([])
#             if i == j:
#                 inv_l[i][j] = 1 / l[i][j]
#                 inv_u[i][j] = 1 / u[i][j]
#             elif i > j:
#                 inv_u[i][j] = 0
#             elif j > i:
#                 inv_l[i][j] = 0
#
#     for j in range(1, n):
#         for i in range(n - 1):
#             if i + j > n - 1:
#                 break
#             else:
#                 inv_u[i][i + j] = -1 * sum([u[i][k] * inv_u[k][i + j] for k in range(i + 1, i + j + 1)]) / u[i][i]
#             if i + j > n - 1:
#                 break
#             else:
#                 inv_l[i + j][i] = -1 * sum([l[i + j][k] * inv_l[k][i] for k in range(i, i + j)]) / l[i + j][i + j]
#
#     # inv_xTx = inv_u * inv_l
#     inv_xtx: List[List[Any]] = list()
#     for i in range(n):
#         inv_xtx.append([])
#         for j in range(n):
#             inv_xtx[i].append([])
#             inv_xtx[i][j] = sum([inv_u[i][k] * inv_l[k][j] for k in range(n)])
#     # pow(inv_xtx[0][0], 0.5) is the errF in Excel Linest function
#
#     # calculate Y values base on the fitted formula
#     estimate_y = [sum([xlst[j][i] * beta[j] for j in range(n)]) for i in range(m)]
#     resid = [(estimate_y[i] - a0[i]) ** 2 for i in range(m)]
#     reg = [(i - sum(estimate_y) / len(estimate_y)) ** 2 for i in estimate_y]
#     ssresid = sum(resid)  # residual sum of squares / sum squared residual
#     ssreg = sum(reg)  # regression sum of square
#     sstotal = ssreg + ssresid  # total sum of squares
#     df = m - n + 1 - 1  # df = degree of freedom
#     m_ssresid = ssresid / df
#     se_beta = [pow(m_ssresid * inv_xtx[i][i], 0.5) for i in range(n)]
#     rseb = (se_beta[0] / beta[0]) * 100 if beta[0] != 0 else se_beta[0]  # relative error of intercept
#     r2 = ssreg / sstotal if sstotal != 0 else 1  # r2 = ssreg / sstotal
#
#     def get_adjusted_y(*args):
#         args = [[1] * len(args[0]), *args]
#         return [sum([beta[i] * args[i][j] for i in range(len(beta))]) for j in range(len(args[0]))]
#
#     return beta[0], se_beta[0], rseb, r2, 'mswd', beta[1:], se_beta[1:], get_adjusted_y, m_ssresid


def quadratic(a0: list, a1: list):
    """ y = b + m1 * x + m2 * x ^ 2
    :param a0: known_y's, y = b + m1 * x + m2 * x ^ 2
    :param a1: known_x's
    :return: intercept | standard error | relative error | r2 | MSWD | [m1, m2] | [sem1, sem2], equation
    """
    # y = b + m1 * x + m2 * x ^ 2
    k = list(linest(a0, a1, [i ** 2 for i in a1]))
    b, seb, rseb, r2, mswd, [b, m1, m2], [seb, sem1, sem2] = k[0:7]

    def get_adjusted_y(x: list):
        return [b + m1 * _x + m2 * _x ** 2 for _x in x]

    k[7] = get_adjusted_y

    return k


def polynomial(a0: list, a1: list, degree: int = 5):
    """ y = b + m1 * x + m2 * x ^ 2 + ... + m[n] * x ^ n
    :param a0: known_y's, y = b + m1 * x + m2 * x ^ 2
    :param a1: known_x's
    :param degree: the order of the fitting, default = 5
    :return: intercept | standard error | relative error | r2 | MSWD | [m1, m2] | [sem1, sem2], equation
    """
    # y = b + m1 * x + m2 * x ^ 2 + ... + m[n] * x ^ n
    k = list(linest(a0, *[[j ** (i + 1) for j in a1] for i in range(degree)]))
    b, seb, rseb, r2, mswd, beta, se_beta = k[0:7]

    def get_adjusted_y(x: list):
        return [sum([beta[i] * _x ** i for i in range(degree + 1)]) for _x in x]

    k[7] = get_adjusted_y

    return k


def logest(a0: list, a1: list):
    """
    :param a0: known_y's, y = b * m ^ x
    :param a1: known_x's
    :return: intercept | standard error | relative error | R2 | MSWD | m | sem
    """
    # y = b * m ^ x, Microsoft Excel LOGEST function, ln(y) = ln(b) + ln(m) * x
    a0 = [np.log(i) for i in a0]  # ln(y)
    linest_res = linest(a0, a1)
    b, seb, rseb, r2, mswd, [b, lnm], [seb, selnm] = linest_res[0:7]
    b = np.exp(b)
    m = np.exp(lnm)
    sem = np.exp(lnm) * selnm
    seb = b * seb  # Excel.Logest function do not consider the error propagation
    rseb = seb / b * 100
    return b, seb, rseb, r2, mswd, m, sem


def power(a0: list, a1: list):
    """
    :param a0: known_y's, y = a * x ^ b + c
    :param a1: known_x's
    :return: intercept | standard error of intercept | relative error | R2 | MSWD | [a, b, c] | [sem, sec, seb]
    """

    def _pow_func(x, a, b, c):
        return a * x ** b + c

    def _solve_pow(params):
        a, b, c = params
        x, y = [0, 0, 0], [0, 0, 0]
        x[0] = sum(a1[:3]) / 3
        y[0] = sum(a0[:3]) / 3
        x[1] = sum(a1) / len(a1)
        y[1] = sum(a0) / len(a0)
        x[2] = sum(a1[-3:]) / 3
        y[2] = sum(a0[-3:]) / 3
        return np.array([
            _pow_func(x[0], a, b, c) - y[0],
            _pow_func(x[1], a, b, c) - y[1],
            _pow_func(x[2], a, b, c) - y[2],
        ])

    def _get_sum(a, b, c):
        y_predicted = [_pow_func(_x, a, b, c) for _x in a1]
        return sum([(y_predicted[i] - a0[i]) ** 2 for i in range(len(a0))])

    def _get_abc(b):  # Return a, b, c given b based on linest regression
        f = linest(a0, [_x ** b for _x in a1])
        return f[5][1], b, f[0]

    try:
        a, b, c = fsolve(func=_solve_pow, x0=np.array([1, 1, 1]))  # initial estimate
        count = 0
        step = 0.01
        while count < 100:
            a, b, c = _get_abc(b)
            s = _get_sum(a, b, c)
            b_left, b_right = b - step * b, b + step * b
            s_left = _get_sum(*_get_abc(b_left))
            s_right = _get_sum(*_get_abc(b_right))
            if s_left > s > s_right:
                b = b_right
                continue
            elif s_left < s < s_right:
                b = b_left
                continue
            elif s_left < s_right:
                b = (b + b_left) / 2
            else:
                b = (b + b_right) / 2
            step = step * 0.5
            count += 1
            if step < 0.000001:
                break
    except RuntimeError:
        raise RuntimeError
    except np.linalg.LinAlgError:
        raise np.linalg.LinAlgError
    except TypeError or IndexError:
        raise IndexError

    f = linest(a0, [_x ** b for _x in a1])
    a, sea, c, sec = f[5][1], f[6][1], f[0], f[1]

    calculated_y = [_pow_func(i, a, b, c) for i in a1]
    resid = [(calculated_y[i] - a0[i]) ** 2 for i in range(len(a0))]
    reg = [(i - sum(calculated_y) / len(calculated_y)) ** 2 for i in calculated_y]
    ssresid = sum(resid)
    ssreg = sum(reg)
    sstotal = ssreg + ssresid
    df = len(a0) - 1
    m_ssresid = ssresid / df
    r2 = ssreg / sstotal if sstotal != 0 else 1

    intercept = c
    se_intercept_1 = sec
    dp = len(a1)  # data points
    z = [i ** b for i in a1]
    # calculate error of intercept
    errfz = pow(sum([i ** 2 for i in z]) / (dp * sum([i ** 2 for i in z]) - sum(z) ** 2), 0.5)
    errfx = pow(sum([i ** 2 for i in a1]) / (dp * sum([i ** 2 for i in a1]) - sum(a1) ** 2), 0.5)
    # seb = errfz * sey = errfz * ssresid / df -> se_intercept = sey * errfx = seb / errfz * errfx
    se_intercept = sec / errfz * errfx
    rse_intercept = se_intercept / intercept * 100

    return intercept, se_intercept, rse_intercept, r2, 'mswd', [a, b, c], 'se', \
           lambda x: [_pow_func(i, a, b, c) for i in x], m_ssresid


def exponential(a0: list, a1: list):
    """
    :param a0: known_y's, y = a * b ^ x + c
    :param a1: known_x's
    :return: intercept | standard error of intercept | relative error | R2 | MSWD | [m, c, b] | [sem, sec, seb]
    """

    def _exp_func(x, a, b, c):
        return a * b ** x + c

    def _solve_exp(params):
        a, b, c = params
        x, y = [0, 0, 0], [0, 0, 0]
        x[0] = sum(a1[:3]) / 3
        y[0] = sum(a0[:3]) / 3
        x[1] = sum(a1) / len(a1)
        y[1] = sum(a0) / len(a0)
        x[2] = sum(a1[-3:]) / 3
        y[2] = sum(a0[-3:]) / 3
        return np.array([
            _exp_func(x[0], a, b, c) - y[0],
            _exp_func(x[1], a, b, c) - y[1],
            _exp_func(x[2], a, b, c) - y[2],
        ])

    def _get_sum(a, b, c):
        y_predicted = [_exp_func(_x, a, b, c) for _x in a1]
        return sum([(y_predicted[i] - a0[i]) ** 2 for i in range(len(a0))])

    def _get_ac(b):
        f = linest(a0, [b ** _x for _x in a1])
        return f[5][1], b, f[0]

    try:
        a, b, c = fsolve(_solve_exp, np.array([1, 1, 1]))
        count = 0
        step = 0.01
        while count < 100:
            a, b, c = _get_ac(b)
            s = _get_sum(a, b, c)
            b_left, b_right = b - step * b, b + step * b
            s_left = _get_sum(*_get_ac(b_left))
            s_right = _get_sum(*_get_ac(b_right))
            if s_left > s > s_right:
                b = b_right
                continue
            elif s_left < s < s_right:
                b = b_left
                continue
            elif s_left < s_right:
                b = (b + b_left) / 2
            else:
                b = (b + b_right) / 2
            count += 1
            step = step * 0.5
            if step < 0.000001:
                break

    except RuntimeError:
        raise RuntimeError
    except np.linalg.LinAlgError:
        raise np.linalg.LinAlgError
    except TypeError:
        raise TypeError
    except IndexError:
        raise IndexError

    f = linest(a0, [b ** _x for _x in a1])
    a, sea, c, sec = f[5][1], f[6][1], f[0], f[1]

    calculated_y = [_exp_func(i, a, b, c) for i in a1]
    resid = [(calculated_y[i] - a0[i]) ** 2 for i in range(len(a0))]
    reg = [(i - sum(calculated_y) / len(calculated_y)) ** 2 for i in calculated_y]
    ssresid = sum(resid)
    ssreg = sum(reg)
    sstotal = ssreg + ssresid
    dp = len(a1)
    df = dp - 1
    m_ssresid = ssresid / df
    r2 = ssreg / sstotal if sstotal != 0 else 1

    z = [b ** i for i in a1]
    intercept = a + c
    se_intercept_1 = np.sqrt(sea ** 2 + sec ** 2)
    # calculate error of intercept
    errfz = pow(sum([i ** 2 for i in z]) / (dp * sum([i ** 2 for i in z]) - sum(z) ** 2), 0.5)
    errfx = pow(sum([i ** 2 for i in a1]) / (dp * sum([i ** 2 for i in a1]) - sum(a1) ** 2), 0.5)
    # seb = errfz * sey = errfz * ssresid / df -> se_intercept = sey * errfx = seb / errfz * errfx
    se_intercept = sec / errfz * errfx
    rse_intercept = se_intercept / intercept * 100

    return intercept, se_intercept, rse_intercept, r2, 'mswd', [a, b, c], 'se', \
           lambda x: [_exp_func(i, a, b, c) for i in x], m_ssresid


""" line functions """


def linear_eq(x: list, beta: list):
    """ y = b0 * x^0 + b1 * x^1 + ... + bn * x^n
    Parameters
    ----------
    beta : coefficients
    x :

    Returns
    -------

    """
    return [np.prod([beta, [_x ** i for i in range(len(beta))]], axis=0).sum() for _x in x]


def exponential_eq(x: list, beta: list):
    """ y = a * b ^ x + c
    Parameters
    ----------
    beta : coefficients, [a, b, c]
    x :

    Returns
    -------

    """
    return [beta[0] * beta[1] ** _x + beta[0] for _x in x]


def power_eq(x: list, beta: list):
    """ y = y = a * x ^ b + c
    Parameters
    ----------
    beta : coefficients, [a, b, c]
    x :

    Returns
    -------

    """
    return [beta[0] * _x ** beta[1] + beta[0] for _x in x]


if __name__ == '__main__':
    pass

